@@ -42,7 +42,7 @@
 # include <sys/ioctl.h>
 #endif
 #if defined(PTYS_ARE_PTMX) && !defined(__CYGWIN32__)
-# include <sys/stropts.h>	/* for I_PUSH */
+# include <sys/stropts.h>       /* for I_PUSH */
 #endif
 
 
@@ -102,10 +102,10 @@ ptytty_get_pty(int *fd_tty, const char **ttydev)
     pfd = open("/dev/ptmx", O_RWNN, 0);
 #  endif
     if (pfd >= 0) {
-      if (grantpt(pfd) == 0	/* change slave permissions */
-	  && unlockpt(pfd) == 0) {	/* slave now unlocked */
-	*ttydev = ptsname(pfd);	/* get slave's name */
-	return pfd;
+      if (grantpt(pfd) == 0     /* change slave permissions */
+          && unlockpt(pfd) == 0) {      /* slave now unlocked */
+        *ttydev = ptsname(pfd); /* get slave's name */
+        return pfd;
       }
       close(pfd);
     }
@@ -140,15 +140,15 @@ ptytty_get_pty(int *fd_tty, const char **ttydev)
       snprintf1(c1, strlen(c1), "%d", idx);
       snprintf1(c2, strlen(c2), "%d", idx);
       if (access(tty_name, F_OK) < 0) {
-	idx = 256;
-	break;
+        idx = 256;
+        break;
       }
       if ((pfd = open(pty_name, O_RWNN, 0)) >= 0) {
-	if (access(tty_name, R_OK | W_OK) == 0) {
-	  *ttydev = strdup(tty_name);
-	  return pfd;
-	}
-	close(pfd);
+        if (access(tty_name, R_OK | W_OK) == 0) {
+          *ttydev = strdup(tty_name);
+          return pfd;
+        }
+        close(pfd);
       }
     }
   }
@@ -160,24 +160,24 @@ ptytty_get_pty(int *fd_tty, const char **ttydev)
     char tty_name[] = "/dev/tty??";
 
 # ifndef PTYCHAR1
-#  define PTYCHAR1	"pqrstuvwxyz"
+#  define PTYCHAR1      "pqrstuvwxyz"
 # endif
 # ifndef PTYCHAR2
-#  define PTYCHAR2	"0123456789abcdef"
+#  define PTYCHAR2      "0123456789abcdef"
 # endif
     for (c1 = PTYCHAR1; *c1; c1++) {
       pty_name[(sizeof(pty_name) - 3)] =
-	tty_name[(sizeof(pty_name) - 3)] = *c1;
+        tty_name[(sizeof(pty_name) - 3)] = *c1;
       for (c2 = PTYCHAR2; *c2; c2++) {
-	pty_name[(sizeof(pty_name) - 2)] =
-	  tty_name[(sizeof(pty_name) - 2)] = *c2;
-	if ((pfd = open(pty_name, O_RWNN, 0)) >= 0) {
-	  if (access(tty_name, R_OK | W_OK) == 0) {
-	    *ttydev = strdup(tty_name);
-	    return pfd;
-	  }
-	  close(pfd);
-	}
+        pty_name[(sizeof(pty_name) - 2)] =
+          tty_name[(sizeof(pty_name) - 2)] = *c2;
+        if ((pfd = open(pty_name, O_RWNN, 0)) >= 0) {
+          if (access(tty_name, R_OK | W_OK) == 0) {
+            *ttydev = strdup(tty_name);
+            return pfd;
+          }
+          close(pfd);
+        }
       }
     }
   }
@@ -201,104 +201,113 @@ ptytty_get_tty(const char *ttydev)
  * Make our tty a controlling tty so that /dev/tty points to us
  */
 /* EXTPROTO */
+
+
 int
 ptytty_control_tty(int fd_tty, const char *ttydev)
 {
-#ifndef __QNX__
-  int fd;
+
+  
 
   DPRINTF3(DEBUG_TERMIO, "pid: %d, tty fd: %d, dev: %s", getpid(), fd_tty,
-	   ttydev);
-/* ---------------------------------------- */
+           ttydev);
   DPRINTF2(DEBUG_TERMIO, "tcgetpgrp(): %d  getpgrp(): %d", tcgetpgrp(fd_tty),
-	   getpgrp());
+           getpgrp());
 
+  /* ------------------- Become leader of our own session:   --------------------- */
 # ifdef HAVE_SETSID
   {
     pid_t ret = setsid();
 
     DPRINTF2(DEBUG_TERMIO, "setsid() returned %d %s", (int)ret,
-	     ERRMSG(ret < 0));
+             ERRMSG(ret < 0));
   }
 # endif
 
-/* ---------------------------------------- */
-# ifdef TIOCNOTTY
-  fd = open("/dev/tty", O_RDWR | O_NOCTTY);
-  DPRINTF1(DEBUG_TERMIO, "Voiding tty associations: previous=%s",
-	   fd < 0 ? "no" : "yes");
-  if (fd >= 0) {
-    int ret = ioctl(fd, TIOCNOTTY, NULL);	/* void tty associations */
-
-    DPRINTF2(DEBUG_TERMIO, "ioctl(..., TIOCNOTTY): %d %s", ret,
-	     ERRMSG(ret < 0));
-    close(fd);
-  }
-# endif
-/* ---------------------------------------- */
-  fd = open("/dev/tty", O_RDWR | O_NOCTTY);
-  DPRINTF1(DEBUG_TERMIO,
-	   "ptytty_control_tty(): /dev/tty has controlling tty? %s",
-	   fd < 0 ? "no (good)" : "yes (bad)");
-  if (fd >= 0)
-    close(fd);			/* ouch: still have controlling tty */
-  errno = 0;
-  /* ---------------------------------------- */
-#if defined(PTYS_ARE_PTMX) && defined(I_PUSH)
-/*
- * Push STREAMS modules:
- *    ptem: pseudo-terminal hardware emulation module.
- *    ldterm: standard terminal line discipline.
- *    ttcompat: V7, 4BSD and XENIX STREAMS compatibility module.
- *
- * After we push the STREAMS modules, the first open() on the slave side
- * (i.e. the next section between the dashes giving us "tty opened OK")
- * should make the "ptem" (or "ldterm" depending upon either which OS
- * version or which set of manual pages you have) module give us a
- * controlling terminal.  We must already have close()d the master side
- * fd in this child process before we push STREAMS modules on because the
- * documentation is really unclear about whether it is any close() on
- * the master side or the last close() - i.e. a proper STREAMS dismantling
- * close() - on the master side which causes a hang up to be sent
- * through - Geoff Wing
- */
-# ifdef HAVE_ISASTREAM
-  if (isastream(fd_tty) == 1)
-# endif
+  /* ------------------- Get controlling tty                 --------------------- */
+#ifndef __QNX__  /* in QNX, the only way to get a controlling tty is at process creation time, using
+                    qnx_spawn() instead of fork(). I'm too lazy to re-write my_pty_fork(), so I let rlwrap
+                    soldier on without a controlling tty */ 
   {
-    DPRINTF0(DEBUG_TERMIO, "Pushing STREAMS modules");
-    ioctl(fd_tty, I_PUSH, "ptem");
-    ioctl(fd_tty, I_PUSH, "ldterm");
-    ioctl(fd_tty, I_PUSH, "ttcompat");
-  }
-#endif
-/* ---------------------------------------- */
-# if defined(TIOCSCTTY)
-  fd = ioctl(fd_tty, TIOCSCTTY, NULL);
-  DPRINTF2(DEBUG_TERMIO, "ioctl(..,TIOCSCTTY): %d %s", fd, ERRMSG(fd < 0));
-# elif defined(TIOCSETCTTY)
-  fd = ioctl(fd_tty, TIOCSETCTTY, NULL);
-  DPRINTF2(DEBUG_TERMIO, "ioctl(..,TIOCSETCTTY): %d %s", fd, ERRMSG(fd < 0));
-# else
-  fd = open(ttydev, O_RDWR);
-  DPRINTF2(DEBUG_TERMIO, "tty open%s %s", (fd < 0 ? " failure" : "ed OK"),
-	   ERRMSG(fd < 0));
-  if (fd >= 0)
-    close(fd);
-# endif
-/* ---------------------------------------- */
-  fd = open("/dev/tty", O_WRONLY);
-  DPRINTF2(DEBUG_TERMIO, "do we have controlling tty now: %s %s",
-	   (fd < 0 ? "no (fatal)" : "yes (good)"), ERRMSG(fd < 0));
-  if (fd < 0)
-    mywarn("Could not get controlling terminal for %s", program_name);	/* mywarn called from child */
-  else
-    close(fd);
+    int fd;
+    
+# ifdef TIOCNOTTY
+    fd = open("/dev/tty", O_RDWR | O_NOCTTY);
+    DPRINTF1(DEBUG_TERMIO, "Voiding tty associations: previous=%s",
+             fd < 0 ? "no" : "yes");
+    if (fd >= 0) {
+      int ret = ioctl(fd, TIOCNOTTY, NULL);       /* void tty associations */
 
-/* ---------------------------------------- */
-  DPRINTF2(DEBUG_TERMIO, "tcgetpgrp(): %d  getpgrp(): %d", tcgetpgrp(fd_tty),
-	   getpgrp());
-/* ---------------------------------------- */
+      DPRINTF2(DEBUG_TERMIO, "ioctl(..., TIOCNOTTY): %d %s", ret,
+               ERRMSG(ret < 0));
+      close(fd);
+    }
+# endif
+    /* ---------------------------------------- */
+    fd = open("/dev/tty", O_RDWR | O_NOCTTY);
+    DPRINTF1(DEBUG_TERMIO,
+             "ptytty_control_tty(): /dev/tty has controlling tty? %s",
+             fd < 0 ? "no (good)" : "yes (bad)");
+    if (fd >= 0)
+      close(fd);                  /* ouch: still have controlling tty */
+    errno = 0;
+    /* ---------------------------------------- */
+#if defined(PTYS_ARE_PTMX) && defined(I_PUSH)
+    /*
+     * Push STREAMS modules:
+     *    ptem: pseudo-terminal hardware emulation module.
+     *    ldterm: standard terminal line discipline.
+     *    ttcompat: V7, 4BSD and XENIX STREAMS compatibility module.
+     *
+     * After we push the STREAMS modules, the first open() on the slave side
+     * (i.e. the next section between the dashes giving us "tty opened OK")
+     * should make the "ptem" (or "ldterm" depending upon either which OS
+     * version or which set of manual pages you have) module give us a
+     * controlling terminal.  We must already have close()d the master side
+     * fd in this child process before we push STREAMS modules on because the
+     * documentation is really unclear about whether it is any close() on
+     * the master side or the last close() - i.e. a proper STREAMS dismantling
+     * close() - on the master side which causes a hang up to be sent
+     * through - Geoff Wing
+     */
+# ifdef HAVE_ISASTREAM
+    if (isastream(fd_tty) == 1)
+# endif
+      {
+        DPRINTF0(DEBUG_TERMIO, "Pushing STREAMS modules");
+        ioctl(fd_tty, I_PUSH, "ptem");
+        ioctl(fd_tty, I_PUSH, "ldterm");
+        ioctl(fd_tty, I_PUSH, "ttcompat");
+      }
+#endif
+    /* ---------------------------------------- */
+# if defined(TIOCSCTTY)
+    fd = ioctl(fd_tty, TIOCSCTTY, NULL);
+    DPRINTF2(DEBUG_TERMIO, "ioctl(..,TIOCSCTTY): %d %s", fd, ERRMSG(fd < 0));
+# elif defined(TIOCSETCTTY)
+    fd = ioctl(fd_tty, TIOCSETCTTY, NULL);
+    DPRINTF2(DEBUG_TERMIO, "ioctl(..,TIOCSETCTTY): %d %s", fd, ERRMSG(fd < 0));
+# else
+    fd = open(ttydev, O_RDWR);
+    DPRINTF2(DEBUG_TERMIO, "tty open%s %s", (fd < 0 ? " failure" : "ed OK"),
+             ERRMSG(fd < 0));
+    if (fd >= 0)
+      close(fd);
+# endif
+    /* ---------------------------------------- */
+    fd = open("/dev/tty", O_WRONLY);
+    DPRINTF2(DEBUG_TERMIO, "do we have controlling tty now: %s %s",
+             (fd < 0 ? "no (fatal)" : "yes (good)"), ERRMSG(fd < 0));
+    if (fd < 0)
+      mywarn("Could not get controlling terminal for %s", program_name);  /* mywarn called from child */
+    else
+      close(fd);
+
+    /* ---------------------------------------- */
+    DPRINTF2(DEBUG_TERMIO, "tcgetpgrp(): %d  getpgrp(): %d", tcgetpgrp(fd_tty),
+             getpgrp());
+    /* ---------------------------------------- */
+  }
 #endif /* ! __QNX__ */
   return 0;
 }
