@@ -16,81 +16,85 @@
     the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
     You may contact the author by:
-       e-mail:  hlub@knoware.nl
+    e-mail:  hlub@knoware.nl
 */
 
 
 #include "rlwrap.h"
 
 /* global vars */
-int remember_for_completion = FALSE;	/* whether we should put al words from in/output on the list */
-char *multiline_separator = NULL;	/* character sequence to use in lieu of newline when storing multi-line input in a single history line */
+int remember_for_completion = FALSE;    /* whether we should put al words from in/output on the list */
+char *multiline_separator = NULL;       /* character sequence to use in lieu of newline when storing multi-line input in a single history line */
 char *pre_given = NULL;                 /* pre-given user input when rlwrap starts up */
-struct rl_state saved_rl_state = { "", "", 0, 0 };	/* saved state of readline */
-static char return_key;	                /* Key pressed to enter line */
+struct rl_state saved_rl_state = { "", "", 0, 0 };      /* saved state of readline */
+static char return_key;                 /* Key pressed to enter line */
 static int forget_line;
 static char *colour_start, *colour_end;        /* colour codes */
 int multiline_prompts = TRUE;
 
 /* forward declarations */
-void line_handler(char *);
+static void line_handler(char *);
 static void my_add_history(char *);
 static int my_accept_line(int, int);
 static int my_accept_line_and_forget(int, int);
 static int dump_all_keybindings(int,int);
 static int munge_line_in_editor(int, int);
+static int please_update_alaf(int,int);
+static int please_update_ce(int,int);
+
+static Keymap getmap(const char *name);
+static void bindkey(int key, rl_command_func_t *function, const char *maplist);
 
 void
 init_readline(char *prompt)
 {
   DPRINTF1(DEBUG_READLINE, "Initialising readline version %x", rl_readline_version);   
-  rl_add_defun("rlwrap_accept_line", my_accept_line, -1);
-  rl_bind_key('\n', my_accept_line);
-  rl_bind_key('\r', my_accept_line); 
+  rl_add_defun("rlwrap-accept-line", my_accept_line,-1);
+  rl_add_defun("rlwrap-accept-line-and-forget", my_accept_line_and_forget,-1);
+  rl_add_defun("rlwrap-dump-all-keybindings", dump_all_keybindings,-1);
+  rl_add_defun("rlwrap-call-editor", munge_line_in_editor, -1);
 
-  rl_add_defun("insert_close", rl_insert_close, -1);
-  rl_bind_key(')', rl_insert_close);
-  rl_bind_key('}', rl_insert_close);
-  rl_bind_key(']', rl_insert_close);
-
-  rl_add_defun("rlwrap_accept_line_and_forget", my_accept_line_and_forget, -1);
-  rl_bind_key(15, my_accept_line_and_forget);	/* ascii #15 (Control-O) is unused in readline's emacs and vi keymaps */
-
-  rl_add_defun("rlwrap_dump_all_keybindings", dump_all_keybindings, -1);
-  /* rl_bind_key(8, dump_all_keybindings); */
+  rl_add_defun("rlwrap_accept_line_and_forget", please_update_alaf,-1);
+  rl_add_defun("rlwrap_call_editor", please_update_ce,-1);
   
-  if (multiline_separator != NULL) {
-    rl_add_defun("rlwrap_call_editor", munge_line_in_editor, -1);
-    rl_bind_key(30, munge_line_in_editor);	/* ascii #30 (Control-^) is unused in readline's emacs and vi keymaps */
-  }
+  rl_variable_bind("blink-matching-paren","on"); /* Shouldn't this be on by default? */
+
+  bindkey('\n', my_accept_line, "emacs-standard; vi-insert; vi-command"); 
+  bindkey('\r', my_accept_line, "emacs-standard; vi-insert; vi-command"); 
+  bindkey(15, my_accept_line_and_forget, "emacs-standard; vi-insert; vi-command");	/* ascii #15 (Control-O) is unused in readline's emacs and vi keymaps */
+  if (multiline_separator) 
+    bindkey(30, munge_line_in_editor, "emacs-standard;vi-insert;vi-command");            /* CTRL-^: unused in vi-insert-mode, hardly used in emacs  (doubles arrow-up) */
+  if (debug)
+    bindkey(7, dump_all_keybindings,"emacs-standard; vi-insert; vi-move; vi-command" /* "emacs-ctlx; emacs-meta" */); /* CTRL-G */
+  
+
   /* rl_variable_bind("gnah","gnerp"); It is not possible to create new readline variables (only functions) */
   rl_catch_signals = 0;
   rl_initialize();		/* This has to happen *before* we set rl_redisplay_function, otherwise
 				   readline will not bother to call tgetent(), will be agnostic about terminal
 				   capabilities and hence not be able to honour e.g. a set horizontal-scroll-mode off
 				   in .inputrc */
-  /* printf("gnah = %s\n", rl_variable_value("gnah")); */ 
+  
   using_history();
   rl_redisplay_function = my_redisplay;
   rl_completion_entry_function =
     (rl_compentry_func_t *) & my_completion_function;
-  /* rl_getc_function = &my_rl_getc_function; */
+  
   rl_catch_signals = FALSE;
   rl_catch_sigwinch = FALSE;
   saved_rl_state.input_buffer = mysavestring("");
   saved_rl_state.raw_prompt = mysavestring("");
   saved_rl_state.cooked_prompt = NULL;
-  if (pre_given) {
-    saved_rl_state.input_buffer = mysavestring(pre_given); /* stuff pre-given text into edit buffer */
-    saved_rl_state.point =  strlen(pre_given);
-    DPRINTF0(DEBUG_READLINE, "Starting line edit (because of -P option)");
-    within_line_edit = TRUE;
-    restore_rl_state();
-  }
+  
 }
 
 
-/* save readline internal stae in rl_state, redisplay the prompt
+
+
+
+
+
+/* save readline internal state in rl_state, redisplay the prompt
    (so that client output gets printed at the right place) */
 void
 save_rl_state()
@@ -99,15 +103,15 @@ save_rl_state()
   saved_rl_state.input_buffer = mysavestring(rl_line_buffer);
   /* saved_rl_state.raw_prompt = mysavestring(rl_prompt); */
  
-  saved_rl_state.point = rl_point;	/* and point    */
+  saved_rl_state.point = rl_point;      /* and point    */
   rl_line_buffer[0] = '\0';
   if (saved_rl_state.already_saved)
     return;
   saved_rl_state.already_saved = 1;
-  rl_delete_text(0, rl_end);	/* clear line  (after prompt) */
+  rl_delete_text(0, rl_end);    /* clear line  (after prompt) */
   rl_point = 0;
-  my_redisplay();		/* and redisplay (this time without user input, cf the comments for the line_handler() function below) */
-  rl_callback_handler_remove();	/* restore original terminal settings */
+  my_redisplay();               /* and redisplay (this time without user input, cf the comments for the line_handler() function below) */
+  rl_callback_handler_remove(); /* restore original terminal settings */
   rl_deprep_terminal();
 
 }
@@ -127,98 +131,62 @@ restore_rl_state()
   cook_prompt_if_necessary();
   newprompt =  mark_invisible(saved_rl_state.cooked_prompt); /* bracket (colour) control sequences with \001 and \002 */  
   rl_expand_prompt(newprompt);
-  mirror_slaves_echo_mode();	/* don't show passwords etc */
+  mirror_slaves_echo_mode();    /* don't show passwords etc */
   
   DPRINTF1(DEBUG_READLINE,"newprompt now %s", mangle_string_for_debug_log(newprompt,MANGLE_LENGTH));
   rl_callback_handler_install(newprompt, &line_handler);
   DPRINTF0(DEBUG_AD_HOC, "freeing newprompt");
   free(newprompt);             /* readline docs don't say it, but we can free newprompt now (readline apparently
-				  uses its own copy) */
+                                  uses its own copy) */
   rl_insert_text(saved_rl_state.input_buffer);
   rl_point = saved_rl_state.point;
   saved_rl_state.already_saved = 0;
   DPRINTF0(DEBUG_AD_HOC, "Starting redisplay");
-  rl_redisplay();
+  rl_redisplay(); 
   rl_prep_terminal(1);
-  we_still_have_to_display_the_prompt =  FALSE; /* has been done right now */
+  prompt_is_still_uncooked =  FALSE; /* has been done right now */
 }
 
-void
+static void
 line_handler(char *line)
 {
   char *rewritten_line, *filtered_line;
-  int i;
   
-  if (line == NULL) {		/* EOF on input, forward it  */
+  if (line == NULL) {           /* EOF on input, forward it  */
     DPRINTF1(DEBUG_READLINE, "EOF detected, writing character %d", term_eof);
     /* colour_the_prompt = FALSE; don't mess with the cruft that may come out of dying command @@@ but command may not die!*/
     write_EOF_to_master_pty();
   } else {
     if (*line &&                 /* forget empty lines  */
-	redisplay &&             /* forget passwords    */
-	!forget_line &&          /* forget lines entered by CTRL-O */
-	!should_be_forgotten(line)) {	/* forget lines matching -g option regexp */ 
+        redisplay &&             /* forget passwords    */
+        !forget_line &&          /* forget lines entered by CTRL-O */
+        !match_regexp(line, forget_regexp, TRUE)) {     /* forget lines (case-inseitively) matching -g option regexp */ 
       my_add_history(line);
     }
     forget_line = FALSE; /* until CTRL-O is used again */
 
-    /* Time for a design decision: when we send the accepted line to
-       the client command, it will most probably be echoed back. We
-       have two choices: we leave the entered line on screen and
-       suppress just enough of the clients output (I believe this is
-       what rlfe does), or we remove the entered input (but not the
-       prompt!) and let it be replaced by the echo.
-
-       This is what we do; as a bonus, if the program doesn't echo,
-       e.g. at a password prompt, the **** which has been put
-       there by our homegrown_redisplay function won't be touched.
-
-       I think this method is more natural for multi-line input as
-       well, (we will actually see our multi-line input as multiple
-       lines) but not everyone will agree with that.
-       
-       The only small annoyance is that vi-mode will have put us on the next
-       line when line_handler is called (is this documented
-       somewhere?) rlfe's client-suppression method won't have any
-       difficulty with this, but rlwrap will have to move the cursor back
-       up. Not all terminals can do this.     
-    */
+    /* Time for a design decision: when we send the accepted line to the client command, it will most probably be echoed
+       back. We have two choices: we leave the entered line on screen and suppress just enough of the clients output (I
+       believe this is what rlfe does), or we remove the entered input (but not the prompt!) and let it be replaced by
+       the echo.
 
+       This is what we do; as a bonus, if the program doesn't echo, e.g. at a password prompt, the **** which has been
+       put there by our homegrown_redisplay function will be removed (@@@is this what we want?)
 
+       I think this method is more natural for multi-line input as well, (we will actually see our multi-line input as
+       multiple lines) but not everyone will agree with that.
+          
+       O.K, we know for sure that cursor is at start of line. When clients output arrives, it will be printed at
+       just the right place - but first we 'll erase the user input (as it may be about to be changed by the filter) */
     
-    /* emacs-mode has the cursor 1 past end of prompt (at start of
-       line).  vi-mode has it on start of new line (Is all of this
-       documented somewhere?)  when in vi-mode, we put the cursor back
-       at the same place it would have been in emacs-mode: */
-
-    
-    if (rl_editing_mode == 0) {  /* i.e. vi mode */
-      char *prompt_plus_line = add2strings(rl_prompt,line);
-      int lines_below_start_of_prompt =
-        1 +  (prompt_is_single_line() ?
-              0 :
-              colourless_strlen(prompt_plus_line, NULL) / winsize.ws_col);
-      free(prompt_plus_line);
-      for (i=0; i < lines_below_start_of_prompt; i++)
-        curs_up();
-      my_putstr(saved_rl_state.cooked_prompt ? saved_rl_state.cooked_prompt : saved_rl_state.raw_prompt); /* @@@ we realy want a rlwrap_visible_prompt() function */
-    }
-
-    
-    /* O.K, we now know for sure that cursor is at start of line. When
-       clients output arrives, it will be printed at just the right
-       place */
-
-    /* but first we 'll erase the user input (as it may be changed by the filter */
-    
-    rl_delete_text(0, rl_end);	/* clear line  (after prompt) */
+    rl_delete_text(0, rl_end);  /* clear line  (after prompt) */
     rl_point = 0;
-    my_redisplay();		/* and redisplay (this time without user input, cf the comments for the line_handler() function below) */
+    my_redisplay();             /* and redisplay (this time without user input, cf the comments for the line_handler() function below) */
 
     rewritten_line =
       (multiline_separator ? 
        search_and_replace(multiline_separator, "\n", line, 0, NULL,
-			  NULL) : mysavestring(line));
+                          NULL) : mysavestring(line));
 
 
     
@@ -228,9 +196,9 @@ line_handler(char *line)
     else { /* password, none of filters business */
       pass_through_filter(TAG_INPUT, "***"); /* filter some input anyway, to keep filter in sync (result is discarded).  */
       filtered_line = mysavestring(rewritten_line);
-    }	
+    }   
     free(rewritten_line);
-	
+        
     
     /* do we have to adapt clients winzsize now? */
     if (deferred_adapt_commands_window_size) {
@@ -244,13 +212,13 @@ line_handler(char *line)
     /*OK, now feed line to underlying command and wait for the echo to come back */
     put_in_output_queue(filtered_line);
     DPRINTF2(DEBUG_READLINE, "putting %d bytes %s in output queue", (int) strlen(rewritten_line),
-	     mangle_string_for_debug_log(rewritten_line, 40));
+             mangle_string_for_debug_log(rewritten_line, 40));
     write_EOL_to_master_pty(return_key ? &return_key : "\n");
-    /* if (one_shot_rlwrap)
-       write_EOF_to_master_pty(); */
-    free_foreign(line);         /* free_foreign because line was malloc'ed by readline, not by rlwrap */
-    free(filtered_line);	/* we're done with them  */
 
+    accepted_lines++;
+    free_foreign(line);         /* free_foreign because line was malloc'ed by readline, not by rlwrap */
+    free(filtered_line);        /* we're done with them  */
+        
     return_key = 0;
     within_line_edit = FALSE;
 
@@ -290,7 +258,6 @@ my_accept_line_and_forget(int count, int key)
   return my_accept_line(count, '\n');
 }
 
-
 static int
 dump_all_keybindings(int count, int key)
 {
@@ -298,12 +265,12 @@ dump_all_keybindings(int count, int key)
   rl_variable_dumper(FALSE);
   rl_macro_dumper(FALSE);
   return 0;
-}	
+}       
 
 /* format line and add it to history list, avoiding duplicates if necessary */
 static void
 my_add_history(char *line)
-{	
+{       
   int lookback, count, here;
   char *new_entry, *filtered_line;
 
@@ -320,7 +287,7 @@ my_add_history(char *line)
   }
 
   
-  new_entry = filtered_line;	
+  new_entry = filtered_line;    
   
   lookback = min(history_length, lookback);      
   for (count = 0, here = history_length + history_base - 1;
@@ -329,7 +296,7 @@ my_add_history(char *line)
     /* DPRINTF3(DEBUG_READLINE, "strcmping <%s> and <%s> (count = %d)", line, history_get(here)->line ,count); */
     if (strncmp(new_entry, history_get(here) -> line, 10000) == 0) {
       HIST_ENTRY *entry = remove_history (here - history_base); /* according to the history library doc this should be
-								   remove_history(here) @@@?*/
+                                                                   remove_history(here) @@@?*/
       DPRINTF2(DEBUG_READLINE, "removing duplicate entry #%d (%s)", here, entry->line);
       free_foreign(entry->line);
       free_foreign(entry);
@@ -339,14 +306,6 @@ my_add_history(char *line)
   free(new_entry);
 }
 
-
-
-
-
-
-
-
-
 /* Homegrown redisplay function - erases current line and prints the
    new one.  Used for passwords (where we want to show **** instead of
    user input) and whenever HOMEGROWN_REDISPLAY is defined (for
@@ -361,19 +320,17 @@ my_add_history(char *line)
 static void
 my_homegrown_redisplay(int hide_passwords)
 {
-  
-
-  static int line_start = 0;	/* at which position of prompt_plus_line does the printed line start? */
+  static int line_start = 0;    /* at which position of prompt_plus_line does the printed line start? */
   static int line_extends_right = 0;
   static int line_extends_left = 0;
   static char *previous_line = NULL;
   
   
   int width = winsize.ws_col;
-  int skip = max(1, min(width / 5, 10));	/* jumpscroll this many positions when cursor reaches edge of terminal */
+  int skip = max(1, min(width / 5, 10));        /* jumpscroll this many positions when cursor reaches edge of terminal */
   
   char *prompt_without_ignore_markers;
-  int colourless_promptlen = colourless_strlen(rl_prompt, &prompt_without_ignore_markers);
+  int colourless_promptlen = colourless_strlen(rl_prompt, &prompt_without_ignore_markers,0);
   int promptlen = strlen(prompt_without_ignore_markers);
   int invisible_chars_in_prompt = promptlen - colourless_promptlen;
   char *prompt_plus_line = add2strings(prompt_without_ignore_markers, rl_line_buffer);
@@ -381,9 +338,9 @@ my_homegrown_redisplay(int hide_passwords)
   int total_length = strlen(prompt_plus_line);
   int curpos = promptlen + rl_point; /* cursor position within prompt_plus_line */
   int i, printed_length,
-      new_curpos,                    /* cursor position on screen */
+    new_curpos,                    /* cursor position on screen */
     keep_old_line, vlinestart, printwidth, last_column;
-  DPRINTF3(DEBUG_AD_HOC,"rl_prompt: <%s>, prompt_without_ignore_markers: <%s>,  prompt_plus_line: <%s>", rl_prompt, prompt_without_ignore_markers, prompt_plus_line); 	
+  DPRINTF3(DEBUG_AD_HOC,"rl_prompt: <%s>, prompt_without_ignore_markers: <%s>,  prompt_plus_line: <%s>", rl_prompt, prompt_without_ignore_markers, prompt_plus_line);   
 
   /* In order to handle prompt with colour we either print the whole prompt, or start past it:
      starting in the middle is too difficult (i.e. I am too lazy) to get it right.
@@ -402,26 +359,26 @@ my_homegrown_redisplay(int hide_passwords)
   
   if (hide_passwords)
     for (i = promptlen; i < total_length; i++)
-      prompt_plus_line[i] = '*';	/* hide a pasword by making user input unreadable  */
+      prompt_plus_line[i] = '*';        /* hide a pasword by making user input unreadable  */
 
 
-  if (rl_point == 0)		/* (re)set  at program start and after accept_line (where rl_point is zeroed) */
+  if (rl_point == 0)            /* (re)set  at program start and after accept_line (where rl_point is zeroed) */
     line_start = 0;
   assert(line_start == 0 || line_start >= promptlen); /* the line *never* starts in the middle of the prompt (too complicated to handle)*/
   vlinestart = (line_start > promptlen ? line_start : invisible_chars_in_prompt); 
   
 
-  if (curpos - vlinestart > width - line_extends_right)	/* cursor falls off right edge ?   */
-    vlinestart = (curpos - width + line_extends_right) + skip;	/* jumpscroll left                 */
+  if (curpos - vlinestart > width - line_extends_right) /* cursor falls off right edge ?   */
+    vlinestart = (curpos - width + line_extends_right) + skip;  /* jumpscroll left                 */
 
-  else if (curpos < vlinestart + line_extends_left) {	/* cursor falls off left edge ?    */
-    if (curpos == total_length)	/* .. but still at end of line?    */
-      vlinestart = max(0, total_length - width);	/* .. try to display entire line   */
-    else			/* in not at end of line ..        */
+  else if (curpos < vlinestart + line_extends_left) {   /* cursor falls off left edge ?    */
+    if (curpos == total_length) /* .. but still at end of line?    */
+      vlinestart = max(0, total_length - width);        /* .. try to display entire line   */
+    else                        /* in not at end of line ..        */
       vlinestart = curpos - line_extends_left - skip; /* ... jumpscroll right ..         */
-  }	
+  }     
   if (vlinestart <= invisible_chars_in_prompt) {
-    line_start = 0;		/* ... but not past start of line! */
+    line_start = 0;             /* ... but not past start of line! */
     vlinestart = invisible_chars_in_prompt;
   } else if (vlinestart > invisible_chars_in_prompt && vlinestart <= promptlen) {
     line_start = vlinestart = promptlen;
@@ -430,7 +387,7 @@ my_homegrown_redisplay(int hide_passwords)
   }
 
   printwidth = (line_start > 0 ? width : width + invisible_chars_in_prompt);
-  printed_length = min(printwidth, total_length - line_start);	/* never print more than width     */
+  printed_length = min(printwidth, total_length - line_start);  /* never print more than width     */
   last_column = printed_length - vlinestart;
 
 
@@ -439,7 +396,7 @@ my_homegrown_redisplay(int hide_passwords)
 
   assert(0 <= line_start);
   assert(line_start <= curpos);
-  assert(curpos <= line_start + printed_length);	/* <=, rather than <, as cursor may be past eol   */
+  assert(curpos <= line_start + printed_length);        /* <=, rather than <, as cursor may be past eol   */
   assert(line_start + printed_length <= total_length);
 
 
@@ -457,7 +414,7 @@ my_homegrown_redisplay(int hide_passwords)
   if (line_extends_right)
     new_line[printwidth - 1] = '>';
 
-  DPRINTF0(DEBUG_AD_HOC, "In the middle...");
+  
 
   keep_old_line = FALSE;
   if (term_cursor_hpos) {
@@ -465,23 +422,23 @@ my_homegrown_redisplay(int hide_passwords)
       keep_old_line = TRUE;
     } else {
       if (previous_line)
-	free(previous_line);
+        free(previous_line);
       previous_line = mysavestring(new_line);
     }
   }
-
-
+  /* DPRINTF2(DEBUG_AD_HOC, "keep_old_line=%d, new_line=<%s>", keep_old_line, new_line); */
+  /* keep_old_line = TRUE; */
   if (!keep_old_line) {
     clear_line();
     cr();
     write_patiently(STDOUT_FILENO, new_line, printed_length, "to stdout");
   }
   
-  assert(term_cursor_hpos || !keep_old_line);	/* if we cannot position cursor, we must have reprinted ... */
+  assert(term_cursor_hpos || !keep_old_line);   /* if we cannot position cursor, we must have reprinted ... */
 
   if (term_cursor_hpos)
     cursor_hpos(new_curpos);
-  else				/* ... so we know we're 1 past last position on line */
+  else                          /* ... so we know we're 1 past last position on line */
     backspace(last_column - new_curpos);
   free(prompt_plus_line);
   free(prompt_without_ignore_markers);
@@ -521,10 +478,10 @@ munge_line_in_editor(int count, int key)
   char *editor_command1, *editor_command2, *editor_command3, *editor_command4,
     *line_number_as_string, *column_number_as_string;
   char *input, *rewritten_input, *rewritten_input2, **possible_tmpdirs, **possible_editor_commands;
-  
-#ifndef HAVE_SYSTEM
-  return 0;    /* we could write our own system() using execv(), but we are too lazy to do that */
-#else
+
+
+  if (!multiline_separator)
+    return 0;
 
   possible_tmpdirs = list4(getenv("TMPDIR"), getenv("TMP"), getenv("TEMP"), "/tmp");
   possible_editor_commands = list4(getenv("RLWRAP_EDITOR"), getenv("EDITOR"), getenv("VISUAL"), "vi +%L");
@@ -547,7 +504,7 @@ munge_line_in_editor(int count, int key)
     myerror("could not create temporary file %s", tmpfilename);
   text_to_edit =
     search_and_replace(multiline_separator, "\n", rl_line_buffer, rl_point,
-		       &line_number, &column_number);
+                       &line_number, &column_number);
   write_patiently(tmpfile_fd, text_to_edit, strlen(text_to_edit), "to temporary file");
 
   if (close(tmpfile_fd) != 0) /* improbable */
@@ -560,16 +517,18 @@ munge_line_in_editor(int count, int key)
   column_number_as_string = as_string(column_number);
   editor_command2 =
     search_and_replace("%L", line_number_as_string, editor_command1, 0, NULL,
-		       NULL);
+                       NULL);
   editor_command3 =
     search_and_replace("%C", column_number_as_string, editor_command2, 0,
-		       NULL, NULL);
+                       NULL, NULL);
   editor_command4 = add3strings(editor_command3, " ", tmpfilename);
 
-  /* call editor, temporarily restoring terminal settings */
-  clear_line();
-  cr();
-  if (terminal_settings_saved && (tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_terminal_settings) < 0))	/* reset terminal */
+  
+  
+  
+
+  /* call editor, temporarily restoring terminal settings */    
+  if (terminal_settings_saved && (tcsetattr(STDIN_FILENO, TCSAFLUSH, &saved_terminal_settings) < 0))    /* reset terminal */
     myerror("tcsetattr error on stdin");
   DPRINTF1(DEBUG_READLINE, "calling %s", editor_command4);
   if ((ret = system(editor_command4))) {
@@ -593,23 +552,27 @@ munge_line_in_editor(int count, int key)
   if (bytes_read < 0)
     myerror("unreadable temp file %s", tmpfilename);
   input[bytes_read] = '\0';
-  rewritten_input = search_and_replace("\t", "    ", input, 0, NULL, NULL);	/* rlwrap cannot handle tabs in input lines */
+  rewritten_input = search_and_replace("\t", "    ", input, 0, NULL, NULL);     /* rlwrap cannot handle tabs in input lines */
   rewritten_input2 =
     search_and_replace("\n", multiline_separator, rewritten_input, 0, NULL,
-		       NULL);
+                       NULL);
   for(p = rewritten_input2; *p ;p++)
     if(*p < ' ')
       *p = ' ';        /* replace all control characters (like \r) by spaces */
-  
+
+
   rl_delete_text(0, strlen(rl_line_buffer));
-  rl_point = 0;
+  rl_point = 0;  
+  clear_line();
+  cr();
+  my_putstr(saved_rl_state.cooked_prompt);
   rl_insert_text(rewritten_input2);
-  rl_point = 0;			/* leave cursor on predictable place */
-  rl_done = 1;
-  return_key = (char)'\n';
+  rl_point = 0;                 /* leave cursor on predictable place */
+  rl_done = 1;                  /* accept line immediately */
 
   
 
+
   /* wash those dishes */
   if (unlink(tmpfilename))
     myerror("could not delete temporary file %s", tmpfilename);
@@ -623,9 +586,9 @@ munge_line_in_editor(int count, int key)
   free(input);
   free(rewritten_input);
   free(rewritten_input2);
-
+  
+  return_key = (char)'\n';
   return 0;
-#endif /* HAVE_SYSTEM */
 }
 
 
@@ -634,6 +597,7 @@ void
 initialise_colour_codes(char *colour)
 {
   int attributes, foreground, background;
+  DPRINTF1(DEBUG_READLINE, "initialise_colour_codes(\"%s\")", colour);
   attributes = foreground = -1;
   background = 40; /* don't need to specify background; 40 passes the test automatically */
   sscanf(colour, "%d;%d;%d", &attributes, &foreground, &background);
@@ -641,8 +605,8 @@ initialise_colour_codes(char *colour)
 #define OUTSIDE(lo,hi,val) (val < lo || val > hi) 
   if (OUTSIDE(0,8,attributes) || OUTSIDE(30,37,foreground) || OUTSIDE(40,47,background))
     myerror("\n"
-	    "  prompt colour spec should be <attr>;<fg>[;<bg>]\n"
-	    "  where <attr> ranges over [0...8], <fg> over [30...37] and <bg> over [40...47]\n"
+            "  prompt colour spec should be <attr>;<fg>[;<bg>]\n"
+            "  where <attr> ranges over [0...8], <fg> over [30...37] and <bg> over [40...47]\n"
             "  example: 0;33 for yellow on current background, 1;31;40 for bold red on black ");
   colour_start= add3strings("\033[", colour,"m");
   colour_end  = "\033[0m";
@@ -674,13 +638,14 @@ move_cursor_to_start_of_prompt(int erase)
 {
   int termwidth = winsize.ws_col;
   int promptlen_on_screen, number_of_lines_in_prompt, curpos, count;
-
-
-
-  if (we_still_have_to_display_the_prompt && ! impatient_prompt)
+  int cooked = (saved_rl_state.cooked_prompt != NULL);
+  
+  DPRINTF2(DEBUG_READLINE,"prompt_is_still_uncooked: %d, impatient_prompt: %d", prompt_is_still_uncooked, impatient_prompt);
+  if (prompt_is_still_uncooked && ! impatient_prompt)
     return; /* @@@ is this necessary ?*/
 
-  promptlen_on_screen =  colourless_strlen_unmarked(saved_rl_state.cooked_prompt ? saved_rl_state.cooked_prompt : saved_rl_state.raw_prompt);
+	
+  promptlen_on_screen =  colourless_strlen_unmarked(saved_rl_state.cooked_prompt ? saved_rl_state.cooked_prompt : saved_rl_state.raw_prompt, termwidth);
   curpos = (within_line_edit ? 1 : 0); /* if the user has pressed a key the cursor will be 1 past the current prompt */ 
   number_of_lines_in_prompt = 1 +  ((promptlen_on_screen + curpos -1) / termwidth); /* integer arithmetic! (e.g. 171/80 = 2) */
   cr(); 
@@ -690,8 +655,8 @@ move_cursor_to_start_of_prompt(int erase)
     curs_up();
   } 
   clear_line();
-  DPRINTF4(DEBUG_READLINE,"moved cursor up %d lines (erase = %d, len=%d, termwidth=%d)", number_of_lines_in_prompt - 1, erase, promptlen_on_screen, termwidth); 
-}	
+  DPRINTF4(DEBUG_READLINE,"moved cursor up %d lines (erase = %d, len=%d, cooked=%d)", number_of_lines_in_prompt - 1, erase, promptlen_on_screen, cooked); 
+}       
 
 
 int
@@ -741,17 +706,17 @@ char *process_new_output(const char* buffer, struct rl_state* state) {
     if (remember_for_completion) {
       feed_line_into_completion_list(result); /* feed output into completion list *after* filtering */
     }
-  } else {	
+  } else {      
     new_prompt = mysavestring(old_prompt_plus_new_output);
     result = mysavestring("");
-  }		    
+  }                 
   free(old_prompt_plus_new_output);
 
   saved_rl_state.raw_prompt = new_prompt;
   if (saved_rl_state.cooked_prompt) {
     free (saved_rl_state.cooked_prompt);
     saved_rl_state.cooked_prompt = NULL; 
-  }	
+  }     
   return result;
 }
 
@@ -759,18 +724,25 @@ char *process_new_output(const char* buffer, struct rl_state* state) {
 int cook_prompt_if_necessary () {
   char *pre_cooked, *filtered, *uncoloured;
   
+  filtered = NULL;
   if (saved_rl_state.cooked_prompt)
     return FALSE;  /* cooked already */
   pre_cooked = mysavestring(saved_rl_state.raw_prompt);
   unbackspace(pre_cooked);
-  filtered = pass_through_filter(TAG_PROMPT, pre_cooked);
+  if ((prompt_regexp && ! match_regexp(pre_cooked, prompt_regexp, FALSE)) ||  /* raw prompt doesn't match '--only-cook' regexp */
+       (strcmp((filtered =  pass_through_filter(TAG_PROMPT, pre_cooked)), "_THIS_CANNOT_BE_A_PROMPT_")== 0)) { /* filter has "refused" the prompt */
+    saved_rl_state.cooked_prompt =  (impatient_prompt ? mysavestring(pre_cooked) : mysavestring("")); /* don't cook, eat raw (and eat nothing if patient) */
+    free(pre_cooked);
+    free(filtered); /* free(NULL) is not an error */
+    return FALSE;
+  }     
   free(pre_cooked);
   if(substitute_prompt) {
     uncoloured = mysavestring(substitute_prompt); 
     free(filtered);
   } else {
     uncoloured = filtered;
-  }	
+  }     
   if (colour_the_prompt) { 
     saved_rl_state.cooked_prompt =  colourise(uncoloured);
     free(uncoloured);
@@ -778,4 +750,51 @@ int cook_prompt_if_necessary () {
     saved_rl_state.cooked_prompt = uncoloured;
   }
   return TRUE;
-}	
+}       
+
+
+/* Utility functions for binding keys. */
+
+static int please_update(const char *varname) {
+  myerror("since version 0.35, the readline function '%s' is called '%s'\n"
+          "(hyphens instead of underscores). Please update your .inputrc",
+          varname, search_and_replace("_","-", varname, 0, NULL, NULL));
+  return 0;
+}       
+
+static int please_update_alaf(int count, int key) {
+  return please_update("rlwrap_accept_line_and_forget");
+}
+
+static int please_update_ce(int count, int key) {
+  return please_update("rlwrap_call_editor");
+}
+
+
+
+/*
+   From the readline documentation: Acceptable keymap names are emacs, emacs-standard, emacs-meta, emacs-ctlx, vi,
+   vi-move, vi-command, and vi-insert. vi is equivalent to vi-command; emacs is equivalent to emacs-standard. The
+   default value is emacs. The value of the editing-mode variable also affects the default keymap.
+*/
+
+
+static Keymap getmap(const char *name) {
+  Keymap km = rl_get_keymap_by_name(name);
+  if (!km) 
+    myerror("Could not get keymap '%s'", name);
+  return km;
+}
+
+static void bindkey(int key, rl_command_func_t *function, const char *maplist) {
+  char *mapnames[] = {"emacs-standard","emacs-ctlx","emacs-meta","vi-insert","vi-move","vi-command",NULL};
+  char **mapname;
+  for (mapname = mapnames; *mapname; mapname++) 
+    if(strstr(maplist, *mapname)) {
+      Keymap kmap = getmap(*mapname);
+      DPRINTF4(DEBUG_READLINE,"Binding key %d (%s) in keymap '%s' to <0x%lx>", key, mangle_char_for_debug_log(key,TRUE), *mapname, (long) function);
+      if (rl_bind_key_in_map(key, function, kmap))
+        myerror("Could not bind key %d (%s) in keymap '%s'", key, mangle_char_for_debug_log(key,TRUE), *mapname);
+    }
+}       
+
