@@ -15,7 +15,7 @@
     the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
     You may contact the author by:
-       e-mail:  hlub@knoware.nl
+    e-mail:  hlub@knoware.nl
 */
 
 #include "../config.h"
@@ -32,8 +32,8 @@
 
 #include <sys/stat.h>
 #include <sys/ioctl.h>
+#include <sys/resource.h>
 #include <fcntl.h>
-#include <termios.h>
 #include <unistd.h>
 #include <signal.h>
 #include <stdio.h>
@@ -45,8 +45,9 @@
 
 
 
-#define __USE_XOPEN
-#define __USE_GNU
+/* #define __USE_XOPEN
+   #define __USE_GNU */
+
 #include <stdlib.h>
 
 #include <sched.h>
@@ -71,23 +72,22 @@
 
 
 
-/* We need tgetent and friends. If there is a term.h, expect it there (may well be a ncurses emulation)
-   If not, and there is a ncurses/term.h, expect it in ncurses/term.h . If not, try curses.h as a last resort
-   @@@ configure should check the presence of tgetent in those header files */
-
+#ifdef HAVE_CURSES_H
+#  include <curses.h>
 #  ifdef HAVE_TERM_H
 #    include <term.h>
 #  else 
 #    ifdef HAVE_NCURSES_TERM_H /* cygwin? AIX? */
 #      include <ncurses/term.h>
-#    else
-#       ifdef HAVE_CURSES_H  /* last-ditch */
-#            include <curses.h> /
-#       endif
 #    endif
 #  endif
+#else
+#  ifdef HAVE_TERMCAP_H
+#     include <termcap.h>
+#  endif
+#endif
 
-
+#include <termios.h>
 
 
 #ifdef HAVE_REGEX_H
@@ -121,22 +121,22 @@ char *strchr(), *strrchr();
 
 
 
-#define BUFFSIZE 40960
+#define BUFFSIZE 512
 
 #ifndef MAXPATHLEN
 #define MAXPATHLEN 512
 #endif
 
 
-#ifdef  HAVE_SNPRINTF		/* don't rely on the compiler understanding variadic macros */
-# define snprintf0(buf,bufsize,format)    		snprintf(buf,bufsize,format)
-# define snprintf1(buf,bufsize,format,arg1) 		snprintf(buf,bufsize,format,arg1)
-# define snprintf2(buf,bufsize,format,arg1,arg2) 	snprintf(buf,bufsize,format,arg1,arg2)
+#ifdef  HAVE_SNPRINTF           /* don't rely on the compiler understanding variadic macros */
+# define snprintf0(buf,bufsize,format)                  snprintf(buf,bufsize,format)
+# define snprintf1(buf,bufsize,format,arg1)             snprintf(buf,bufsize,format,arg1)
+# define snprintf2(buf,bufsize,format,arg1,arg2)        snprintf(buf,bufsize,format,arg1,arg2)
 #else
-# define snprintf0(buf,bufsize,format)      		sprintf(buf,format)
-# define snprintf1(buf,bufsize,format,arg1) 		sprintf(buf,format,arg1)
-# define snprintf2(buf,bufsize,format,arg1,arg2) 	sprintf(buf,format,arg1,arg2)
-# define vsnprintf(buf,bufsize,format,ap)   		vsprintf(buf,format,ap)
+# define snprintf0(buf,bufsize,format)                  sprintf(buf,format)
+# define snprintf1(buf,bufsize,format,arg1)             sprintf(buf,format,arg1)
+# define snprintf2(buf,bufsize,format,arg1,arg2)        sprintf(buf,format,arg1,arg2)
+# define vsnprintf(buf,bufsize,format,ap)               vsprintf(buf,format,ap)
 #endif
 
 
@@ -158,15 +158,17 @@ char *strchr(), *strrchr();
 #  define rl_readline_version 0xbaddef
 #endif
 
-#ifdef SPY_ON_READLINE
-# ifndef HOMEGROWN_REDISPLAY
+#if defined(SPY_ON_READLINE)
+extern int _rl_eof_char; /* Spying on readline's private life .... */
+extern int _rl_horizontal_scroll_mode;  
+# if !defined(HOMEGROWN_REDISPLAY)
 #  define MAYBE_MULTILINE 1
 # endif
+#else
+# define _rl_eof_char 0;
 #endif
 
 #ifdef MAYBE_MULTILINE
-extern int _rl_horizontal_scroll_mode;	/* Spying on readline's private life .... */
-
 #  define redisplay_multiple_lines (!_rl_horizontal_scroll_mode)
 #else
 #  define redisplay_multiple_lines (strncmp(rl_variable_value("horizontal-scroll-mode"),"off",3) == 0)
@@ -190,46 +192,48 @@ extern int i_am_child;
 extern int nowarn;
 extern int debug;
 extern char *password_prompt_search_string;
+extern int one_shot_rlwrap;
 extern char *substitute_prompt;
 extern char *history_format;
 extern char *forget_regexp;
+extern char *prompt_regexp;
 extern int renice;
 extern int ignore_queued_input;
 extern int history_duplicate_avoidance_policy;
 extern int pass_on_sigINT_as_sigTERM;
 /* now follow the possible values for history_duplicate_avoidance_policy: */
-#define KEEP_ALL_DOUBLES 		0
-#define ELIMINATE_SUCCESIVE_DOUBLES	1
-#define ELIMINATE_ALL_DOUBLES  		2
+#define KEEP_ALL_DOUBLES                0
+#define ELIMINATE_SUCCESIVE_DOUBLES     1
+#define ELIMINATE_ALL_DOUBLES           2
 extern int one_shot_rlwrap;
 extern int ansi_colour_aware;
 extern int colour_the_prompt;
 extern int received_WINCH;
-extern int we_still_have_to_display_the_prompt;
+extern int prompt_is_still_uncooked;
 extern int impatient_prompt;
-extern int we_just_got_a_signal;
+extern int we_just_got_a_signal_or_EOF;
 extern int remember_for_completion;
 extern int commands_children_not_wrapped; 
-
+extern int accepted_lines;
 
 void cleanup_rlwrap_and_exit(int status);
 void put_in_output_queue(char *stuff);
-
+int  output_queue_is_nonempty(void);
+void flush_output_queue(void);
 
 /* in readline.c: */
 extern struct rl_state
-{				/* struct to save readline state while we're processing output from slave command*/
+{                               /* struct to save readline state while we're processing output from slave command*/
   char *input_buffer;           /* current input buffer */
   char *raw_prompt;             /* current prompt */  
   char *cooked_prompt;          /* ditto redefined by user, or with colour added */
   int point;                    /* cursor position within input buffer */
-  int already_saved;		/* flag set when saved, cleared when restored */
+  int already_saved;            /* flag set when saved, cleared when restored */
 } saved_rl_state;
 
 void save_rl_state(void);
 void restore_rl_state(void);
 void init_readline(char *);
-void line_handler(char *);
 void my_redisplay(void);
 void initialise_colour_codes(char *colour);
 void reprint_prompt(int coloured);
@@ -251,13 +255,15 @@ extern int leave_prompt_alone;
 /* in signals.c */
 extern int command_is_dead;
 extern int commands_exit_status;
+extern int filter_is_dead;
+extern int filters_exit_status;
 extern int sigterm_received;
 extern int deferred_adapt_commands_window_size;
 extern int signal_handlers_were_installed;
 extern int received_sigALRM;
 
 #ifndef RETSIGTYPE
-   #define RETSIGTYPE void /* systems where RETSIGTYPE = int have died out, apparently */
+#define RETSIGTYPE void /* systems where RETSIGTYPE = int have died out, apparently */
 #endif
 typedef RETSIGTYPE (*sighandler_type)(int);       
 
@@ -274,6 +280,7 @@ void suicide_by(int sig, int status);
 int  adapt_tty_winsize(int from_fd, int to_fd);
 void myalarm(int msec);
 void handle_sigALRM(int signo);
+char *signal_name(int signal);
 
 
 /* in utils.c */
@@ -286,6 +293,7 @@ int   write_patiently(int fd, const void *buffer, int count, const char *whither
 void  read_patiently2(int fd, void *buffer, int count, int uninterruptible_msec, const char *whence);
 void  write_patiently2(int fd, const void *buffer, int count, int uninterruptible_msec, const char *whither);
 void  mysetenv(const char *name, const char *value);
+void  set_ulimit(int resource, long value);
 void  usage(int status);
 void  mywarn(const char *message, ...);
 void  myerror(const char *message, ...);
@@ -297,7 +305,8 @@ size_t filesize(const char *filename);
 void  open_logfile(const char *filename);
 void  write_logfile(const char *str);
 void  close_logfile(void);
-void  timestamp(FILE *fp);
+void  timestamp(char *buf, int size);
+int   killed_by(int status);
 void  change_working_directory(void);
 void  log_terminal_settings(struct termios *terminal_settings);
 void  log_fd_info(int fd);
@@ -327,20 +336,21 @@ char *mangle_char_for_debug_log(char c, int quote_me);
 char *mangle_string_for_debug_log(const char *string, int maxlen);
 char *mangle_buffer_for_debug_log(const char *buffer, int length);
 char *search_and_replace(char *patt, char *repl, const char *string,
-			 int cursorpos, int *line, int *col);
+                         int cursorpos, int *line, int *col);
 char *first_of(char **strings);
 char *as_string(int i);
 char *append_and_expand_history_format(char *line);
+void remove_padding_and_terminate(char *buf, int length);
 void unbackspace(char* buf);
+void test_unbackspace (const char *input, const char *expected_result);
 char *mark_invisible(const char *buf);
 char *copy_and_unbackspace(const char *original);
-int colourless_strlen(const char *str, char **copy_without_ignore_markers);
-int colourless_strlen_unmarked (const char *str);
+int colourless_strlen(const char *str, char **pcopy_without_ignore_markers, int termwidth);
+int colourless_strlen_unmarked (const char *str, int termwidth);
 char *get_last_screenline(char *long_line, int termwidth);
 char *lowercase(const char *str);
 char *colour_name_to_ansi_code(const char *colour_name);
-int should_be_forgotten(const char* input_line);
-
+int match_regexp(const char *string, const char *regexp, int case_insensitive);
 
 
 /* in pty.c: */
@@ -372,8 +382,8 @@ char *my_completion_function(char *prefix, int state);
 extern int completion_is_case_sensitive;
 
 
- /* in term.c: */
-int redisplay;			/* TRUE when user input should be readable (instead of *******)  */
+/* in term.c: */
+extern int redisplay;                  /* TRUE when user input should be readable (instead of *******)  */
 void init_terminal(void);
 void set_echo(int);
 void prepare_terminal(void);
@@ -384,7 +394,7 @@ void clear_the_screen(void);
 void curs_up(void);
 void curs_down(void);
 void test_terminal(void);
-int my_putchar(int c);
+int my_putchar(TPUTS_PUTC_ARGTYPE c);
 void my_putstr(const char *string);
 int cursor_hpos(int col);
 extern struct termios saved_terminal_settings;
@@ -392,7 +402,7 @@ extern int terminal_settings_saved;
 extern struct winsize winsize;
 extern char *term_name;
 extern char *term_backspace, term_eof, term_stop, *term_cursor_hpos,
-  *term_cursor_up, *term_cursor_down, *term_newline;
+  *term_cursor_up, *term_cursor_down;
 extern int newline_came_last;
 
 /* in filter.c */
@@ -412,8 +422,8 @@ extern int newline_came_last;
 #define out_of_band(tag) (tag & 128)
 
 
-pid_t filter_pid;
-int filter_is_dead;
+extern pid_t filter_pid;
+extern int filter_is_dead;
 void spawn_filter(const char *filter_command);
 void kill_filter(void);
 char *pass_through_filter(int tag, const char *buffer);
@@ -442,18 +452,23 @@ void filter_test(void);
 #endif
 
 
+
+
 #include "malloc_debug.h" /* malloc_debug.{c,h} not ready for prime time */
 
 #define DEBUG_FILENAME "/tmp/rlwrap.debug"
+#define KA_BOOM  {char *p = (char *) 1; *p = 'c';} /* dump core right here  */ 
+#define KA_SCRUNCH {volatile int x=1, y=0; x = x/y;} /* force a SIGFPE */
+#define KA_SCREECH kill(getpid(),SIGTRAP);        /* enter the debugger - use it to set (conditional) breakpoints from within C code: if (condition) KA_SCREECH; */
 
 /* DPRINTF0 and its ilk  doesn't produce any output except when DEBUG is #defined (via --enable-debug configure option) */
 
 #ifdef  DEBUG
 
 
-#  define DEBUG_TERMIO                     	 1
-#  define DEBUG_SIGNALS                    	 2
-#  define DEBUG_READLINE                   	 4
+#  define DEBUG_TERMIO                           1
+#  define DEBUG_SIGNALS                          2
+#  define DEBUG_READLINE                         4
 #  define DEBUG_MEMORY_MANAGEMENT                8   /* used with malloc_debug.c */
 #  define DEBUG_FILTERING                        16
 #  define DEBUG_AD_HOC                           64  /* only used during rlwrap development */
@@ -461,49 +476,57 @@ void filter_test(void);
 #  define FORCE_HOMEGROWN_REDISPLAY              256 /* force use of my_homegrown_redisplay()        */
 #  define DEBUG_LONG_STRINGS                     512 /* log all strings completely, however long they are */ 
 #  define DEBUG_RACES                            1024 /* introduce random delays */  
-#  define DEBUG_TEST_MAIN                        2048 /* run test_main and exit  */
+#  define DEBUG_RANDOM_FAIL                      2048 /* fail tests randomly */
+#  define DEBUG_TEST_MAIN                        4096 /* run test_main and exit  */
+
+#  define DEBUG_MAX                              DEBUG_TEST_MAIN
 #  define MANGLE_LENGTH                          ((debug_saved & DEBUG_LONG_STRINGS) ? 0 : 20) /* debug_saved is defined within DPRINTF macro */ 
-#  define DEBUG_ALL                        	 (DEBUG_TERMIO | DEBUG_SIGNALS | DEBUG_READLINE | DEBUG_AD_HOC)
-                                                                   
-#  ifdef __GNUC__
-#    define WHERE_AND_WHEN              	  int debug_saved = debug; \
-                                                  if(debug & DEBUG_WITH_TIMESTAMPS) timestamp(debug_fp);\
-                                                  debug = 0; /* don't debug while evaluating the DPRINTF arguments */ \
-                                                  fprintf(debug_fp, "%15s:%-3d\t%-25s ", __FILE__, __LINE__, __FUNCTION__)
-#  else
-#    define WHERE_AND_WHEN                        int debug_saved = debug; \
-                                                  if(debug & DEBUG_WITH_TIMESTAMPS) timestamp(debug_fp);\
-                                                  debug = 0;\
-                                                  fprintf(debug_fp, "%15s:%-3d\t", __FILE__, __LINE__)
-#  endif
+#  define DEBUG_DEFAULT                          (DEBUG_TERMIO | DEBUG_SIGNALS | DEBUG_READLINE)
+#  define DEBUG_ALL                              (2*DEBUG_MAX-1)
 
 
-#  define NL_AND_FLUSH                            fputc('\n', debug_fp) ; fflush(debug_fp); debug = debug_saved;
 
-#  define DPRINTF0(mask, format)\
-  if (debug & mask && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format); NL_AND_FLUSH; }
 
-#  define DPRINTF1(mask, format,arg)\
-  if (debug & mask && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg); NL_AND_FLUSH; }
+# ifndef __GNUC__
+# define __FUNCTION__ ""
+# endif
 
-#  define DPRINTF2(mask, format,arg1, arg2)\
-  if (debug & mask && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2); NL_AND_FLUSH; }
+# define WHERE_AND_WHEN \
+  int debug_saved = debug; char file_line[100], when[100];              \
+  if(debug & DEBUG_WITH_TIMESTAMPS) timestamp(when, sizeof(when)); else *when='\0'; \
+  debug = 0; /* don't debug while evaluating the DPRINTF arguments */ \
+  snprintf2(file_line, sizeof(file_line),"%.15s:%d:",__FILE__,__LINE__); \
+  fprintf(debug_fp, "%-20s %s %-25.25s ", file_line, when, __FUNCTION__);\
 
-#  define DPRINTF3(mask, format,arg1, arg2, arg3)\
-  if (debug & mask && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2, arg3); NL_AND_FLUSH; }
 
-#  define DPRINTF4(mask, format,arg1, arg2, arg3, arg4)\
-  if (debug & mask && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2, arg3,arg4); NL_AND_FLUSH; }
+#  define NL_AND_FLUSH           fputc('\n', debug_fp) ; fflush(debug_fp); debug = debug_saved;
 
-#  define ERRMSG(b)                               (b && (errno != 0) ? add3strings("(", strerror(errno), ")") : "" )
+#  define DPRINTF0(mask, format)                                        \
+  if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format); NL_AND_FLUSH; }
 
-#  define SHOWCURSOR(c)          if (debug & DEBUG_READLINE) {my_putchar(c); mymicrosleep(800); backspace(1);} /* (doesn't work correctly at last column!)*/
+#  define DPRINTF1(mask, format,arg)                                    \
+  if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg); NL_AND_FLUSH; }
 
-#  define DEBUG_RANDOM_SLEEP        if (debug & DEBUG_RACES) {int sleeptime=rand()&127; DPRINTF1(DEBUG_RACES,"going to sleep for %d msecs", sleeptime); mymicrosleep(sleeptime);}
-#  define KA_BOOM  {char *p = 1; *p = 0;} /* dump core right here (too lazy to set breakpoints ....) */ 
+#  define DPRINTF2(mask, format,arg1, arg2)                             \
+  if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2); NL_AND_FLUSH; }
 
-#else 
-#  define MANGLE_LENGTH 0
+#  define DPRINTF3(mask, format,arg1, arg2, arg3)                       \
+  if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2, arg3); NL_AND_FLUSH; }
+
+#  define DPRINTF4(mask, format,arg1, arg2, arg3, arg4)                 \
+  if ((debug & mask) && debug_fp) {WHERE_AND_WHEN; fprintf(debug_fp, format, arg1, arg2, arg3,arg4); NL_AND_FLUSH; }
+
+#  define ERRMSG(b)              (b && (errno != 0) ? add3strings("(", strerror(errno), ")") : "" )
+
+#  define SHOWCURSOR(c)          if (debug & DEBUG_READLINE) {my_putchar(c); mymicrosleep(800); backspace(1);} /* (may work incorrectly at last column!)*/
+
+#  define DEBUG_RANDOM_SLEEP        if (debug & DEBUG_RACES) {int sleeptime=rand()&31; DPRINTF1(DEBUG_RACES,"sleeping for %d msecs", sleeptime); mymicrosleep(sleeptime);}
+
+
+#else
+#  define HEAVEN_FORBID(exp)     (exp)
+#  define OH_PLEASE(exp)         (exp)
+#  define MANGLE_LENGTH          0
 #  define DPRINTF0(mask, format)
 #  define DPRINTF1(mask, format,arg)
 #  define DPRINTF2(mask, format,arg1, arg2)
@@ -511,7 +534,7 @@ void filter_test(void);
 #  define DPRINTF4(mask, format,arg1, arg2, arg3, arg4)
 #  define ERRMSG(b)
 #  define SHOWCURSOR
-#  define NDEBUG		/* disable assertions */
+#  define NDEBUG                /* disable assertions */
 #  define DEBUG_RANDOM_SLEEP  
 #endif
 
