@@ -56,7 +56,7 @@ mystrlcat(char *dst, const char *src, size_t size)
 {
 #ifdef HAVE_STRLCAT
   strlcat(dst, src, size);
-  dst[size - 1] = '\0';		/* we don't check for truncation, just assure '\0'-termination. */
+  dst[size - 1] = '\0';         /* we don't check for truncation, just assure '\0'-termination. */
 #else
   strncat(dst, src, size - strnlen(dst, size) - 1);
   dst[size - 1] = '\0';
@@ -73,7 +73,7 @@ mystrndup(const char *string, int len)
 {
   /* allocate copy of string on the heap */
   char *buf;
-
+  assert(string != NULL);
   buf = (char *)mymalloc(len + 1);
   mystrlcpy(buf, string, len + 1);
   return buf;
@@ -100,8 +100,8 @@ add3strings(const char *str1, const char *str2, const char *str3)
   int size;
   char *buf;
   
-  assert(str1); assert(str2); assert(str3);
-  size = strlen(str1) + strlen(str2) + strlen(str3) + 1;	/* total length plus 0 byte */
+  assert(str1!= NULL); assert(str2!= NULL); assert(str3!= NULL);
+  size = strlen(str1) + strlen(str2) + strlen(str3) + 1;        /* total length plus 0 byte */
   buf = (char *) mymalloc(size);
 
   /* DPRINTF3(DEBUG_TERMIO,"size1: %d, size2: %d, size3: %d",   (int) strlen(str1), (int) strlen(str2),  (int) strlen(str3)); */
@@ -122,7 +122,7 @@ append_and_free_old(char *str1, const char *str2)
     char *result = add2strings(str1,str2);
     free (str1);
     return result;
-  }	
+  }     
 }
 
 
@@ -131,7 +131,7 @@ append_and_free_old(char *str1, const char *str2)
 
 char *
 mybasename(char *filename)
-{				/* determine basename of "filename" */
+{                               /* determine basename of "filename" */
 
   char *p;
 
@@ -144,7 +144,7 @@ mybasename(char *filename)
 
 char *
 mydirname(char *filename)
-{				/* determine directory component of "name" */
+{                               /* determine directory component of "name" */
   char *p;
 
   /* find last '/' in name (if any) */
@@ -164,7 +164,7 @@ char *mystrtok(const char *s, const char *delim) {
     scratchpad = mysavestring(s);
   }    
   return strtok(s ? scratchpad : NULL, delim);
-}	
+}       
 
 /* split_with("a bee    cee"," ") returns a pointer to an array {"a", "bee",  "cee", NULL} on the heap */
 char **split_with(const char *string, const char *delim) {
@@ -175,7 +175,7 @@ char **split_with(const char *string, const char *delim) {
     *pword++ = mysavestring(token);
   *pword = NULL;
   return list;
-}	
+}       
 
 /* split_with("a\t\tbla","\t") returns {"a" "bla", NULL}, but we want {"a", "", "bla", NULL} for filter completion.
    We write a special version (can be freed with free_splitlist) */
@@ -196,7 +196,7 @@ char **split_on_single_char(const char *string, char c) {
   *pword = NULL;
   free(stringcopy);
   return list;
-}	
+}       
 
 /* free_splitlist(list) frees lists components and then list itself. list must be NULL-terminated */
 void free_splitlist (char **list) {
@@ -204,7 +204,7 @@ void free_splitlist (char **list) {
   while(*p)
     free(*p++);
   free (list);
-}	
+}       
 
 
 /* search_and_replace() is a utilty for handling multi-line input
@@ -218,51 +218,55 @@ void free_splitlist (char **list) {
    and 3 in col because a cursor position of 11 in "To be NL ..."
    corresponds to the 3rd column on the 2nd line in "To be \n ..."
 
-   col and line may be NULL pointers.
+   cursorpos, col and line only make sense if repl == "\n", otherwise
+   they may be 0/NULL (search_and_replace works for any repl). The
+   first position on the string corresponds to cursorpos = col = 0 and
+   line = 1.
 */
    
 
 char *
 search_and_replace(char *patt, char *repl, const char *string, int cursorpos,
-		   int *line, int *col)
+                   int *line, int *col)
 {
   int i, j, k;
   int pattlen = strlen(patt);
   int replen = strlen(repl);
   int stringlen = strlen(string);
-  int past_cursorpos = 0;
+  int cursor_found = FALSE;
   int current_line = 1;
-  int current_column = 1;
+  int current_column = 0;
   size_t scratchsize;
   char *scratchpad, *result;
 
-
+  assert(patt && repl && string);
   DPRINTF2(DEBUG_READLINE, "string=%s, cursorpos=%d",
-	   mangle_string_for_debug_log(string, 40), cursorpos);
-  scratchsize = max(stringlen, (stringlen * replen) / pattlen) + 1;	/* worst case : repleng > pattlen and string consists of only <patt> */
+           mangle_string_for_debug_log(string, 40), cursorpos);
+  scratchsize = max(stringlen, (stringlen * replen) / pattlen) + 1;     /* worst case : repleng > pattlen and string consists of only <patt> */
   DPRINTF1(DEBUG_READLINE, "Allocating %d bytes for scratchpad", (int) scratchsize);
   scratchpad = mymalloc(scratchsize);
 
 
-  for (i = j = 0; i < stringlen;) {
-    if (strncmp(patt, string + i, pattlen) == 0) {
-      i += pattlen;
-      for (k = 0; k < replen; k++)
-	scratchpad[j++] = repl[k];
-      current_line++;
-      current_column = 1;
+  for (i = j = 0; i < stringlen;  ) {
+    if (line && col &&                           /* if col and line are BOTH non-NULL, and .. */
+        i >= cursorpos && !cursor_found) {       /*  ... for the first time, i >= cursorpos: */
+      cursor_found = TRUE;                       /* flag that we're done here */                              
+      *line = current_line;                      /* update *line */ 
+      *col = current_column;                     /* update *column */
+    }
+    if (strncmp(patt, string + i, pattlen) == 0) { /* found match */
+      i += pattlen;                                /* update i ("jump over" patt (and, maybe, cursorpos)) */  
+      for (k = 0; k < replen; k++)                 /* append repl to scratchpad */
+        scratchpad[j++] = repl[k];
+      current_line++;                              /* update line # (assuming that repl = "\n") */
+      current_column = 0;                          /* reset column */
     } else {
-      scratchpad[j++] = string[i++];
+      scratchpad[j++] = string[i++];               /* or else just copy things */
       current_column++;
     }
-    if (i >= cursorpos && !past_cursorpos) {
-      past_cursorpos = 1;
-      if (line)
-	*line = current_line;
-      if (col)
-	*col = current_column;
-    }
   }
+  if (line && col)
+    DPRINTF2(DEBUG_READLINE, "line=%d, col=%d", *line, *col);
   scratchpad[j] = '\0';
   result = mysavestring(scratchpad);
   free(scratchpad);
@@ -273,7 +277,7 @@ search_and_replace(char *patt, char *repl, const char *string, int cursorpos,
 /* first_of(&string_array) returns the first non-NULL element of string_array  */
 char *
 first_of(char **strings)
-{				
+{                               
   char **p;
 
   for (p = strings; *p == NULL; p++);
@@ -312,16 +316,16 @@ mangle_char_for_debug_log(char c, int quote_me)
   }
   if (!special) {
     if (c > 0 && c < 27  ) {
-      format = "<CTRL-%c>"; code =  c + 96;
+      format = "<CTRL-%c>"; code =  c + 64;
     } else if (c > 27 && c < 32) {
       format = "<CTRL-%c>"; code =  remainder[c-28];
     } else {
       format = (quote_me ? "\"%c\"" : "%c"); code = c;
-    }	
+    }   
     snprintf1(scrap, sizeof(scrap), format, code);
   }
   return mysavestring (special ? special : scrap);
-}	
+}       
 
 
 /* mangle_string_for_debug_log(string, len) returns a printable
@@ -342,7 +346,7 @@ mangle_string_for_debug_log(const char *string, int maxlen)
     mangled_char = mangle_char_for_debug_log(*p, FALSE);
     total_length +=  strlen(mangled_char);
     if (maxlen && (total_length > maxlen)) {
-      result = append_and_free_old(result, "...");	
+      result = append_and_free_old(result, "...");      
       break;  /* break *before* we reach maxlen */
     }
     result = append_and_free_old(result, mangled_char);
@@ -377,8 +381,8 @@ int scan_metacharacters(const char* string, const char *metacharacters) {
     if (strchr(string, *c))
       return TRUE;
   return FALSE;
-}	
-	       
+}       
+               
 /* allocate and init array of 4 strings (helper for munge_line_in_editor() and filter) */
 char **
 list4 (char *el0, char *el1, char *el2, char *el3)
@@ -392,40 +396,96 @@ list4 (char *el0, char *el1, char *el2, char *el3)
 }
 
 
-	
-char ESCAPE = '\033';
-char BACKSPACE = '\010';
-char CARRIAGE_RETURN = '\015';
-
-static int within_control_seq = FALSE;
+/* remove_padding_and_terminate(buf, N) overwrites buf with a copy of
+   its first N bytes, omitting any zero bytes, and then terminates the
+   result with a final zero byte.  Example: if buf="a\0b\0\0cde@#!" then,
+   after calling remove_padding_and_terminate(buf, 8) buf will contain
+   "abcde\0de@#!"
 
+   We need to call this function on everything we get from the
+   inferior command because (out of sheer programmer laziness) rlwrap
+   uses C strings internally. Zero bytes are only ever used as padding
+   (@@@is this true?), and padding is not used anymore on modern
+   terminals. (except maybe for things like the visual bell) */
+ 
 
+void remove_padding_and_terminate(char *buf, int length) {
+  char *readptr, *copyptr;
 
+  for (readptr = copyptr = buf; readptr < buf + length; readptr++) {
+    if (*readptr != '\0')
+      *copyptr++ = *readptr;
+  }
+  *copyptr = '\0';
+  if (debug && strlen(buf) != length)
+    DPRINTF2(DEBUG_TERMIO, "removed %d zero bytes (padding?) from %s", length - (int) strlen(buf), mangle_string_for_debug_log(buf, MANGLE_LENGTH));
+}       
+        
+#define ESCAPE  '\033'
+#define BACKSPACE '\010'
+#define CARRIAGE_RETURN  '\015'
 
 /* unbackspace(&buf) will overwrite buf (up to and including the first
    '\0') with a copy of itself. Backspaces will move the "copy
    pointer" one backwards, carriage returns will re-set it to the
    begining of buf.  Because the re-written string is always shorter
-   than the original, we need not worry about writing outside buf */
+   than the original, we need not worry about writing outside buf
+
+   Example: if buf="abc\bd\r\e" then, after calling unbackspace(buf),
+   buf will contain "ebd"
+
+   We need this function because many commands emit "status lines"
+   using backspaces and carriage returns to re-write parts of the line in-place.
+   Rlwrap will consider such lines as "prompts" (@@@myabe it shouldn't?)
+   but mayhem results if we feed the \b and \r characters to readline
+*/
 
 void
 unbackspace(char* buf) {
-  char *p, *q;	
-  for(p = q = buf; *p; p++)  {
-    if (*p == BACKSPACE)
-      q = (q > buf ? q - 1 : q);
-    else if (*p == CARRIAGE_RETURN)
-       q = buf; 
-    else {
-      assert (q <= p);
-      assert (q >= buf); 	
-      *q++ = *p;
+  char *readptr, *copyptr, *endptr;
+  int seen_bs_or_cr;
+
+  DPRINTF1(DEBUG_TERMIO,"unbackspace: %s", mangle_string_for_debug_log(buf, MANGLE_LENGTH));
+  seen_bs_or_cr = FALSE;
+  
+  for (readptr = copyptr = endptr = buf; *readptr; readptr++) {
+
+    assert(endptr <= readptr);
+    assert(copyptr <= endptr);
+
+    switch (*readptr) {
+    case BACKSPACE:
+      copyptr = (copyptr > buf ? copyptr - 1 : buf);  /* cannot backspace past beginning of buf */
+      seen_bs_or_cr = TRUE;
+      break;
+    case CARRIAGE_RETURN:
+      copyptr = buf;  
+      seen_bs_or_cr = TRUE;
+      break;
+    default:
+      *copyptr++ = *readptr;      
+      break;
     }
+    if (copyptr > endptr)
+      endptr = copyptr;
   }
-  *q = '\0';
+  *endptr = '\0';
+  if (seen_bs_or_cr) 
+      DPRINTF1(DEBUG_TERMIO,"unbackspace result: %s", mangle_string_for_debug_log(buf, MANGLE_LENGTH));
+  
 }
 
 
+void test_unbackspace (const char *input, const char *expected_result) {
+  char *scrap = mysavestring(input);
+  unbackspace(scrap);
+  if (strcmp(scrap, expected_result) != 0)
+      myerror("unbackspace %s yielded %s, expected %s",
+              mangle_string_for_debug_log(input,0),
+              mangle_string_for_debug_log(scrap,0),
+              expected_result);
+}
+
 
 /* Readline allows to single out character sequences that take up no
    physical screen space when displayed by bracketing them with the
@@ -439,16 +499,17 @@ unbackspace(char* buf) {
 
 /*
   (Re-)definitions for testing   
-#undef RL_PROMPT_START_IGNORE
-#undef  RL_PROMPT_END_IGNORE
-#undef isprint
-#define isprint(c) (c != 'x')
-#define RL_PROMPT_START_IGNORE '('
-#define RL_PROMPT_END_IGNORE ')'
-#define ESCAPE 'E'
+  #undef RL_PROMPT_START_IGNORE
+  #undef  RL_PROMPT_END_IGNORE
+  #undef isprint
+  #define isprint(c) (c != 'x')
+  #define RL_PROMPT_START_IGNORE '('
+  #define RL_PROMPT_END_IGNORE ')'
+  #define ESCAPE 'E'
 */
 
 
+/* TODO @@@ replace the following obscure and unsafe functions using the regex library */
 static void  match_and_copy_ESC_sequence (const char **original, char **copy);
 static void match_and_copy(const char *charlist, const char **original, char **copy);
 static int matches (const char *charlist, char c) ;
@@ -457,23 +518,25 @@ static void copy_next(int n, const char **original, char **copy);
 char *
 mark_invisible(const char *buf)
 {
-    char *scratchpad = mymalloc(3 * strlen(buf) + 1); /* worst case: every char in buf gets surrounded by RL_PROMPT_{START,END}_IGNORE */
-    char *result = scratchpad;
-    const char **original = &buf;
-    char **copy = &scratchpad;
+  int padsize = 3 * strlen(buf) + 1; /* worst case: every char in buf gets surrounded by RL_PROMPT_{START,END}_IGNORE */
+  char *scratchpad = mymalloc (padsize);
+  char *result = scratchpad;
+  const char **original = &buf;
+  char **copy = &scratchpad;
 
-    if (strchr(buf, RL_PROMPT_START_IGNORE))
-      return mysavestring(buf); /* "invisible" parts already marked */
+  if (strchr(buf, RL_PROMPT_START_IGNORE))
+    return mysavestring(buf); /* "invisible" parts already marked */
     
-    while (**original) {
-      /* printf ("orig: %p, copy: %p, result: %s @ %p\n", *original, *copy, result, result); */
-      match_and_copy_ESC_sequence(original, copy);
-      /* match_and_copy_unprintable(original, copy); */
-      copy_next(1, original, copy);
-    }
-    **copy = '\0';
-    return(result);	
-}	
+  while (**original) {
+    /* printf ("orig: %p, copy: %p, result: %s @ %p\n", *original, *copy, result, result); */
+    match_and_copy_ESC_sequence(original, copy);
+    /* match_and_copy_unprintable(original, copy); */
+    copy_next(1, original, copy);
+    assert(*copy - scratchpad < padsize);
+  }
+  **copy = '\0';
+  return(result);       
+}       
 
 
 
@@ -488,14 +551,14 @@ match_and_copy_ESC_sequence (const char **original, char **copy)
   match_and_copy(";0123456789", original, copy);
   match_and_copy("m", original, copy);
   *(*copy)++ = RL_PROMPT_END_IGNORE;
-}	
+}       
     
 static void
 match_and_copy(const char *charlist, const char **original, char **copy)
 {
   while (matches(charlist, **original))
     *(*copy)++ = *(*original)++;
-}	
+}       
 
 static int
 matches (const char *charlist, char c)
@@ -505,7 +568,7 @@ matches (const char *charlist, char c)
     if (*p == c)
       return TRUE;
   return FALSE;
-}	
+}       
   
 
 static void
@@ -532,11 +595,11 @@ copy_and_unbackspace(const char *original)
       copy = copy_start;
     else
       *copy++ = *original;
-  }	
+  }     
   *copy = '\0';
   return copystart;
 #else
-    return copy;
+  return copy;
 #endif
 }
   
@@ -546,41 +609,70 @@ copy_and_unbackspace(const char *original)
    marked with RL_PROMPT_*_IGNORE).  Puts a copy without the
    RL_PROMPT_*_IGNORE characters in *copy_without_ignore_markers (if
    != NULL)
-   */
+*/
 
 int
-colourless_strlen(const char *str, char **copy_without_ignore_markers)
+colourless_strlen(const char *str, char **pcopy_without_ignore_markers, int termwidth)
 {
-  int counting = TRUE, count = 0;
-  const char *p;
-  char *q =  NULL; 
-  if (copy_without_ignore_markers) 
-    q = *copy_without_ignore_markers = mymalloc(strlen(str)+1);
-    
+  int visible = TRUE;
+  int column = 0;
+  int length = strlen(str);
+  char *p, *q, *copy_without_ignore_markers;
+  
+
+  assert(termwidth >= 0);
+  q = copy_without_ignore_markers = mymalloc(length + 1);
+	
   for(p = str; *p; p++) {
+    assert (q < copy_without_ignore_markers + length); 
     switch (*p) {
     case RL_PROMPT_START_IGNORE:
-      counting = FALSE;
+      visible = FALSE;
       continue;
     case RL_PROMPT_END_IGNORE:
-      counting = TRUE;
+      visible = TRUE;
       continue;
-    }	
-    
-    count += (counting ? 1 : 0);
-    if (copy_without_ignore_markers)
+    case '\r':
+      if (visible) {
+        q -= column;
+        column = 0;
+        continue;
+      }
+      break;
+    case '\b':
+      if (visible & q > copy_without_ignore_markers) {
+        q -= 1;
+        column -= 1;
+        if (termwidth && column < 0) 
+          column += termwidth;
+        continue;
+      }
+      break;
+    }   
+    if (visible) {
       *q++ = *p;
-  }
-  if (copy_without_ignore_markers)
-    *q = '\0';
-  return count;
+      column +=1;
+      if (termwidth && column >= termwidth)
+        column -= termwidth;
+    }
+  } 
+  *q = '\0';
+  DPRINTF4(DEBUG_READLINE, "colourless_strlen(\"%s\", 0x%lx, %d) = %d",
+           mangle_string_for_debug_log(str, MANGLE_LENGTH), (long) pcopy_without_ignore_markers,
+           termwidth,  q - copy_without_ignore_markers);  
+  if (pcopy_without_ignore_markers)
+    *pcopy_without_ignore_markers = copy_without_ignore_markers;
+  else
+    free (copy_without_ignore_markers);
+
+  return q - copy_without_ignore_markers;
 }
 
 int
-colourless_strlen_unmarked (const char *str)
+colourless_strlen_unmarked (const char *str, int termwidth)
 {
   char *marked_str = mark_invisible(str);
-  int colourless_length = colourless_strlen(marked_str, NULL);
+  int colourless_length = colourless_strlen(marked_str, NULL, termwidth);
   free(marked_str);
   return colourless_length;
 }
@@ -604,13 +696,13 @@ get_last_screenline(char *long_line, int termwidth)
   
   if (termwidth == 0 ||              /* this may be the case on some weird systems */
       line_length <=  termwidth)  {  /* line doesn't extend beyond right margin
-					@@@ are there terminals that put the cursor on the
-					next line if line_length == termwidth?? */
-     return line_copy; 
+                                        @@@ are there terminals that put the cursor on the
+                                        next line if line_length == termwidth?? */
+    return line_copy; 
   } else if (strchr(long_line, '\033')) { /* <ESC> found, give up */
     free (line_copy);
     return mysavestring("Ehhmm..? > ");
-  } else {	
+  } else {      
     removed = (line_length / termwidth) * termwidth;   /* integer arithmetic: 33/10 = 3 */
     last_screenline  = mysavestring(line_copy + removed);
     free(line_copy);
@@ -626,7 +718,7 @@ lowercase(const char *str) {
   for (p=result; *p; p++)
     *p = tolower(*p);
   return result;
-}	
+}       
 
 
 
@@ -636,7 +728,7 @@ colour_name_to_ansi_code(const char *colour_name) {
     char *lc_colour_name = mysavestring(lowercase(colour_name));
     char *bold_code = (isupper(*colour_name) ? "1" : "0");
 
-    #define isit(c) (strcmp(c,lc_colour_name)==0)
+#define isit(c) (strcmp(c,lc_colour_name)==0)
     char *colour_code =
       isit("black")   ? "30" :
       isit("red")     ? "31" :
@@ -647,49 +739,69 @@ colour_name_to_ansi_code(const char *colour_name) {
       isit("purple")  ? "35" :
       isit("cyan")    ? "36" :
       isit("white")   ? "37" :
-                        NULL ;
+      NULL ;
       
-    #undef isit
+#undef isit
     if (colour_code)
       return add3strings(bold_code,";",colour_code);
     else
-      myerror("unrecognised colour name %s", colour_name);
+      myerror("unrecognised colour name '%s'. Use e.g. 'yellow' or 'Blue'.", colour_name);
   }
   return mysavestring(colour_name);
-}	
+}       
     
     
 
 
-/* match lowercased input_line against forget_regexp (given by the -g option)
-   returns nonzero for successfull match, 0 otherwise */
 
-int
-should_be_forgotten(const char* input_line) {
-  char *error_message;
-  if (!forget_regexp)
+
+/*
+  returns TRUE if 'string' matches the 'regexp' (or is a superstring
+  of it, when we don't HAVE_REGEX_H). The regexp is recompiled with
+  every call, which doesn't really hurt as this function is not called
+  often: at most twice for every prompt.  'string' and 'regexp' may be
+  NULL (in which case FALSE is returned)
+
+  Only used for the --forget-regexp and the --prompt-regexp options
+*/
+int match_regexp (const char *string, const char *regexp, int case_insensitive) {
+  int result;
+  
+  if (!regexp || !string)
     return FALSE;
+  
 #ifndef HAVE_REGEX_H
-  return mystrstr(lowercase(input_line), forget_regexp) ? TRUE : FALSE;
+  {
+    static int been_warned = 0;
+    char *metachars = "*()+?";
+    char *lc_string = (case_insensitive ? lowercase(string) : mysavestring(string));
+    char *lc_regexp = (case_insensitive  ? lowercase(regexp) : mysavestring(regexp));
+
+    if (scan_metacharacters(regexp, metachars) && !been_warned++) /* warn only once if the user specifies a metacharacter */
+      mywarn("one of the regexp metacharacters \"%s\" occurs in regexp(?) \"%s\"\n"
+             "  ...but on your platform, regexp matching is not supported!", metachars, regexp);        
+    
+    result = mystrstr(lc_string, lc_regexp);
+    free(lc_string);
+    free(lc_regexp);    
+  }
 #else
-  { static regex_t *preg = NULL;
-    int retval, size;
-    if (!preg) {
-      preg = mymalloc(sizeof(regex_t));
-      retval = regcomp(preg, forget_regexp, REG_EXTENDED|REG_NOSUB);
+  {
+    regex_t *compiled_regexp = mymalloc(sizeof(regex_t));
+    int compile_error = regcomp(compiled_regexp, regexp, REG_EXTENDED|REG_NOSUB|(case_insensitive ? REG_ICASE : 0));
      
-      if (retval) {
-	size = regerror(retval, preg, NULL, 0);
-	error_message =  mymalloc(size);
-	regerror(retval, preg, error_message, size);
-	myerror("in regex %s: %s", forget_regexp, error_message);
-      }	
+    if (compile_error) {
+      int size = regerror(compile_error, compiled_regexp, NULL, 0);
+      char *error_message =  mymalloc(size);
+      regerror(compile_error, compiled_regexp, error_message, size);
+      errno=0; myerror("in regexp \"%s\": %s", regexp, error_message);  
+    } else {
+      result = !regexec(compiled_regexp, string, 0, NULL, 0);
+      free(compiled_regexp);
     }
-    return !regexec(preg, lowercase(input_line), 0, NULL, 0);
   }
 #endif
+
+  
+  return result;        
 }
-
-
-
-
