@@ -13,7 +13,7 @@
     the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
     You may contact the author by:
-       e-mail:  hlub@knoware.nl
+    e-mail:  hlub@knoware.nl
 */
 
 
@@ -28,90 +28,99 @@ char slaves_working_directory[MAXPATHLEN+1];
 static FILE *log_fp;
 
 
+/* Give up the processor so that other processes (like command) can have their say. Especially useful with multi-line
+   input: if we, after writing a line to the inferior command, would return to the main select() loop immediately, we
+   would find more output to write, and keep blathering away without giving the poor command time to say something
+   back */
 
 void
 yield()
 {
-  #ifdef HAVE_SCHED_YIELD
-    sched_yield();            
-  #else
-    { struct timeval ten_millisecs = { 0, 10000 }; 
-      select(0, NULL, NULL, NULL, &ten_millisecs); /* poor man's sched_yield() */
-    }	
-  #endif
+  DPRINTF0(DEBUG_TERMIO, "yield...");
+#if 0 /* If command writes some output and immediately starts a long computation, sched_yield() will not return for a
+         long time,and rlwrap will not see the output for all that time. The "poor mans sched_yield" with select
+         actually works better! */
+  sched_yield();            
+#else
+  { struct timeval a_few_millisecs = { 0, 1000 }; 
+    select(0, NULL, NULL, NULL, &a_few_millisecs); /* poor man's sched_yield() */
+  }     
+#endif
 }
 
 
-/* 
-
-     Even though even older linux systems HAVE PSELECT, is is non-atomic: signal handlers may (and
-     generally will) run between the unblocking of the signals and the select call (which will
-     then wait untill hell freezes over) Therefore we always convert the contents op
-     ptimeout_ts into to a static struct timeval my_select_timeout_tv, and use the function
-     zero_select_timeout to set this to {0,0} from within a signal handler, which will then
-     make select return immediately.  (Linus Torvalds mentions this trick in 
-     http://lkml.indiana.edu/hypermail/linux/kernel/0506.1/1191.html),
-
-     As i know of no reliable way to distiguish real pselect from fake pselect at configure
-     time HAVE_REAL_PSELECT will be undefined on all systems, even those that do have a genuine
-     pselect.
-     
- */
-
-
-
+static int signal_handled = FALSE;
 
 #ifdef HAVE_REAL_PSELECT
 
 void zero_select_timeout () {
-  ;
+  signal_handled = TRUE;
 }
 
 #else
-static int signal_handled = FALSE;
+
+
+#define until_hell_freezes  NULL ;
+static struct timeval *pmy_select_timeout_tv;
 static struct timeval my_select_timeout_tv;
-static struct timeval until_hell_freezes = {10000000,0};
 
 void zero_select_timeout () {
   my_select_timeout_tv.tv_sec = 0;
   my_select_timeout_tv.tv_usec = 0;
+  pmy_select_timeout_tv = &my_select_timeout_tv;
   signal_handled = TRUE;  
 }
 #endif
 
+/* 
+   Even though even older linux systems HAVE PSELECT, is is non-atomic: signal handlers may (and
+   generally will) run between the unblocking of the signals and the select call (which will
+   then wait untill hell freezes over) Therefore we always convert the contents op
+   ptimeout_ts into to a static struct timeval my_select_timeout_tv, and use the function
+   zero_select_timeout to set this to {0,0} from within a signal handler, which will then
+   make select return immediately.  (Linus Torvalds mentions this trick in 
+   http://lkml.indiana.edu/hypermail/linux/kernel/0506.1/1191.html),
+
+   As I know of no reliable way to distiguish real pselect from fake pselect at configure
+   time HAVE_REAL_PSELECT will be undefined on all systems, even those that do have a genuine
+   pselect.
+*/
+
 
 
 int
 my_pselect(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, const struct timespec *ptimeout_ts, const sigset_t *sigmask) {
-    int retval;
-  
-#ifdef HAVE_REAL_PSELECT
- 
-  return pselect(n, readfds, writefds, exceptfds, &my_pselect_timeout_ts, sigmask);
-#else
-    sigset_t oldmask;
-    if (ptimeout_ts) { /* convert timespec e.g {0,40000000} to timeval {0, 40000} */
-      my_select_timeout_tv.tv_sec = ptimeout_ts -> tv_sec;
-      my_select_timeout_tv.tv_usec = ptimeout_ts -> tv_nsec / 1000;
-    } else
-      my_select_timeout_tv = until_hell_freezes;
-    signal_handled = FALSE;
-    
-    sigprocmask(SIG_SETMASK, sigmask, &oldmask);
-    retval = select(n, readfds, writefds, exceptfds, &my_select_timeout_tv);
-    sigprocmask(SIG_SETMASK, &oldmask, NULL);
 
-    if (signal_handled && retval >= 0) { 
-      errno = EINTR;
-      return -1;
-    } else {
-      return retval;   
-    }
+#ifdef HAVE_REAL_PSELECT
+  return pselect(n, readfds, writefds, exceptfds, ptimeout_ts, sigmask);
+#else
+  int retval;
+  sigset_t oldmask;
+  if (ptimeout_ts) { /* convert timespec e.g {0,40000000} to timeval {0, 40000} */
+    pmy_select_timeout_tv = &my_select_timeout_tv;
+    pmy_select_timeout_tv -> tv_sec = ptimeout_ts -> tv_sec;
+    pmy_select_timeout_tv -> tv_usec = ptimeout_ts -> tv_nsec / 1000;
+  } else
+    pmy_select_timeout_tv = until_hell_freezes;
+  signal_handled = FALSE;
+    
+  sigprocmask(SIG_SETMASK, sigmask, &oldmask);
+  /* most signals will be catched HERE (and their handlers will set my_select_timeout_tv to {0,0}) */
+  retval = select(n, readfds, writefds, exceptfds, pmy_select_timeout_tv);
+  /* but even if they are slow off the mark and get catched HERE the code 3 lines below will notice */
+  sigprocmask(SIG_SETMASK, &oldmask, NULL);
+
+  if (signal_handled && retval >= 0) { 
+    errno = EINTR;
+    return -1;
+  } else {
+    return retval;   
+  }
 #endif
-}	
+}       
 
 struct termios *
-my_tcgetattr(int fd, char *which) /* the which argument is only needed for error reporting */
+my_tcgetattr(int fd, char *which) /* the 'which' argument is only needed for error reporting */
 {
   struct termios *pterm =
     (struct termios *)mymalloc(sizeof(struct termios));
@@ -126,31 +135,15 @@ do_nothing(int unused)
 }
 
 
-int
-read_patiently(int fd, void *buffer, int count, char *whence) {
-  int nread = 0;
-  int total_read =  0;
-  assert (count >= 0);
-  if (count > 0) {
-    while (1) {
-      assert(count > total_read);
-      if((nread = read(fd, (char *) buffer + total_read, count - total_read)) <= 0) {
-	if (errno == EINTR)
-	  continue;
-	else if (nread == 0) { /* EOF, because count - total_read> 0 */
-	  return FALSE;
-	} else 
-	  myerror("read error %s", whence);
-      }
-      total_read += nread;
-      if (total_read == count) /* done */      
-	break;
-    }
-  }	
-  ((char *) buffer)[count] = '\0';
-  return TRUE;
-}      
 
+/* When using read (2) or write (2), one always has to be prepared to handle incomplete
+   reads or writes. The {read,write}_patiently* routines below do just that: they repeat the read
+   or write until the whole buffer has been filled, or an error occurs.
+   The {read,write}_patiently2 functions are specialised for reading/writing the input/output pipe
+   of a filter */
+
+   
+   
 int
 write_patiently(int fd, const void *buffer, int count, const char *whither) {
   int nwritten = 0;
@@ -161,32 +154,35 @@ write_patiently(int fd, const void *buffer, int count, const char *whither) {
   while(1) { 
     if((nwritten = write(fd, (char *)buffer + total_written, count - total_written)) <= 0) {
       if (errno == EINTR)
-	continue;
+        continue;
       else if (errno == EPIPE || nwritten == 0) {
-	return FALSE;
+        return FALSE;
       } else 
-	myerror("write error %s", whither);
-      }	
+        myerror("write error %s", whither);
+    }   
     
     total_written += nwritten;
     if (total_written == count) /* done */      
       break;
-  } 	 
+  }      
   return TRUE;
-}	
+}       
+
+
+
+static int user_frustration_signals[] = {SIGHUP, SIGINT, SIGQUIT, SIGTERM, SIGALRM};
 
 /* keep  reading from fd and write into buffer until count bytes have been read.
    for uninterruptible_msec, restart when interrupted by a signal, after this,
    bail out with an error. if count < 0, returns a 0-terminated copy of buffer  */
 
-static int user_frustration_signals[] = {SIGHUP, SIGINT, SIGQUIT, SIGTERM, SIGALRM};
 
 void
 read_patiently2 (int fd,
-		 void *buffer,
-		 int count,
-		 int uninterruptible_msec,
-		 const char *whence) {
+                 void *buffer,
+                 int count,
+                 int uninterruptible_msec,
+                 const char *whence) {
   int nread = 0;
   int total_read =  0;
   int interruptible = FALSE;
@@ -198,30 +194,30 @@ read_patiently2 (int fd,
     while (1) {
       assert(count > total_read);
       if((nread = read(fd, (char *) buffer + total_read, count - total_read)) <= 0) {
-	if (nread < 0 && errno == EINTR) {
-	  if (interruptible) {
-	    errno = 0; myerror("(user) interrupt, filter_hangs?", whence);
-	  }
-	  if (received_sigALRM) {
-	    received_sigALRM = FALSE;
-	    interruptible = TRUE;
-	  }
-	  continue;
-	} else if (nread == 0)  {
-	  errno = 0; myerror("EOF reading %s", whence);
-	} else { /* nread < 0 */	
-	  myerror("error reading %s",  whence);
-	}
+        if (nread < 0 && errno == EINTR) {
+          if (interruptible) {
+            errno = 0; myerror("(user) interrupt, filter_hangs?", whence);
+          }
+          if (received_sigALRM) {
+            received_sigALRM = FALSE;
+            interruptible = TRUE;
+          }
+          continue;
+        } else if (nread == 0)  {
+          errno = 0; myerror("EOF reading %s", whence);
+        } else { /* nread < 0 */        
+          myerror("error reading %s",  whence);
+        }
       }
       total_read += nread;
       if (total_read == count)  /* done */
-	break;
+        break;
       
     }
-  }	
+  }     
   myalarm(0); /* reset alarm */
   block_all_signals();
-  DPRINTF2(DEBUG_TERMIO, "read_patiently2 read %d bytes: %s", total_read, mangle_buffer_for_debug_log(buffer, total_read));
+  DPRINTF2(DEBUG_AD_HOC, "read_patiently2 read %d bytes: %s", total_read, mangle_buffer_for_debug_log(buffer, total_read));
   return;
 }      
 
@@ -232,10 +228,10 @@ read_patiently2 (int fd,
 
 void
 write_patiently2(int fd,
-		     const void *buffer,
-		     int count,
-		     int uninterruptible_msec,
-		     const char* whither) {
+                 const void *buffer,
+                 int count,
+                 int uninterruptible_msec,
+                 const char* whither) {
   int nwritten = 0;
   int total_written =  0;
   int interruptible = FALSE;
@@ -245,30 +241,30 @@ write_patiently2(int fd,
   if (count == 0)
     return; /* no-op */
   unblock_signals(user_frustration_signals); /* allow impatient users to hit CTRL-C, but only after uninterruptible_msec */
-  myalarm(uninterruptible_msec);	
+  myalarm(uninterruptible_msec);        
   while(1) { 
     if((nwritten = write(fd, (char *)buffer + total_written, count - total_written)) <= 0) {
-	if (nwritten < 0 && errno == EINTR) {
-	  if (interruptible)
-	    errno = 0; myerror("(user) interrupt - filter hangs?");
-	  if (received_sigALRM) {
-	    received_sigALRM = FALSE;
-	    interruptible = TRUE;
-	  }
-	  continue;
-	} else { /* nwritten== 0 or < 0 with error other than EINTR */
-	  myerror("error writing %s", whither);
-	}
+      if (nwritten < 0 && errno == EINTR) {
+        if (interruptible)
+          errno = 0; myerror("(user) interrupt - filter hangs?");
+        if (received_sigALRM) {
+          received_sigALRM = FALSE;
+          interruptible = TRUE;
+        }
+        continue;
+      } else { /* nwritten== 0 or < 0 with error other than EINTR */
+        myerror("error writing %s", whither);
       }
+    }
     total_written += nwritten;
     if (total_written == count) /* done */      
       break;
   }
   myalarm(0);
-  DPRINTF2(DEBUG_TERMIO, "write_patiently2 wrote %d bytes: %s", total_written, mangle_buffer_for_debug_log(buffer, total_written));
+  DPRINTF2(DEBUG_AD_HOC, "write_patiently2 wrote %d bytes: %s", total_written, mangle_buffer_for_debug_log(buffer, total_written));
   block_all_signals();
   return;
-}	
+}       
 
 
 
@@ -278,21 +274,31 @@ mysetenv(const char *name, const char *value)
   int return_value = 0;
   
   
-  #ifdef HAVE_SETENV
-     return_value = setenv(name, value, TRUE);   
-  #elif defined(HAVE_PUTENV)
-     char *name_is_value = add3strings (name, "=", value);
-     return_value = putenv (name_is_value);
-  #else /* won't happen, but anyway: */	   
-     mywarn("setting environment variable %s=%s failed, as this system has neither setenv() nor putenv()", name, value);
-  #endif
+#ifdef HAVE_SETENV
+  return_value = setenv(name, value, TRUE);   
+#elif defined(HAVE_PUTENV)
+  char *name_is_value = add3strings (name, "=", value);
+  return_value = putenv (name_is_value);
+#else /* won't happen, but anyway: */      
+  mywarn("setting environment variable %s=%s failed, as this system has neither setenv() nor putenv()", name, value);
+#endif
      
   if (return_value != 0)
     mywarn("setting environment variable %s=%s failed%s", name, value,
-	   (errno ? "" : " (insufficient environment space?)"));     /* will setenv(...) = -1  set errno? */
+           (errno ? "" : " (insufficient environment space?)"));     /* will setenv(...) = -1  set errno? */
 }       
-	  
-
+          
+void set_ulimit(int resource, long value) {
+  #ifdef HAVE_SETRLIMIT
+  struct rlimit limit;
+  int result;
+  
+  limit.rlim_cur = value;
+  result = setrlimit(resource, &limit);
+  DPRINTF4(DEBUG_ALL, "setrlim() used to set resource #%d to value %ld, result = %d (%s)",
+          resource, value, result, (result == 0 ? "success" : "failure"));
+  #endif
+}
 
 /* private helper function for myerror() and mywarn() */
 static void
@@ -305,23 +311,27 @@ utils_warn(const char *message, va_list ap)
 
   snprintf(buffer, sizeof(buffer) - 1, "%s: ", program_name);
   vsnprintf(buffer + strlen(buffer), sizeof(buffer) - strlen(buffer) - 1,
-	    message, ap);
+            message, ap);
   if (saved_errno)
     snprintf(buffer + strlen(buffer), sizeof(buffer) - strlen(buffer) - 1,
-	     ": %s", strerror(saved_errno));
+             ": %s", strerror(saved_errno));
   mystrlcat(buffer, "\n", sizeof(buffer));
 
   fflush(stdout);
-  fputs(buffer, stderr); /* @@@ error reporting (still) uses bufered I/O */
+  if (nowarn) {
+    DPRINTF1(DEBUG_ALL, "%s", buffer);
+  } else
+    fputs(buffer, stderr); /* @@@ error reporting (still) uses bufered I/O */
+  
   fflush(stderr);
-  DPRINTF1(DEBUG_ALL, "Error/warning: %s", buffer);
+  errno =  saved_errno;
   /* we want this because sometimes error messages (esp. from client) are dropped) */
 
 }
 
 /* myerror("utter failure in %s", where) prints a NL-terminated error
    message ("rlwrap: utter failure in xxxx\n") and exits rlwrap */
-
+  
 
 static char*
 markup(const char*str)
@@ -330,13 +340,13 @@ markup(const char*str)
     return add3strings("\033[1;31m", str,"\033[0m");
   else
     return mysavestring(str);
-}	
+}       
 
 void
 myerror(const char *message, ...)
 {
   va_list ap;
-  char *error_message = add2strings(markup("error: "), message);
+  const char *error_message = message;
 
   va_start(ap, message);
   utils_warn(error_message, ap);
@@ -354,8 +364,7 @@ mywarn(const char *message, ...)
   va_list ap;
   char *warning_message;
 
-  if (nowarn)
-    return;
+  
   warning_message = add2strings(markup("warning: "), message);
   va_start(ap, message);
   utils_warn(warning_message, ap);
@@ -411,24 +420,33 @@ close_open_files_without_writing_buffers() /* called from child just before exec
 
 
 void
-timestamp(FILE *fp)
+timestamp(char *buf, int size)
 {
   struct timeval now;
-  static struct timeval firsttime;
+  static struct timeval firsttime; /* remember when first called */
   static int never_called = 1;
   long diff_usec;
   float diff_sec;
   
   gettimeofday(&now, NULL);
   if (never_called) {
-    firsttime = now;
+    firsttime = now; 
     never_called = 0;
   }
   diff_usec = 1000000 * (now.tv_sec -firsttime.tv_sec) + (now.tv_usec - firsttime.tv_usec);
   diff_sec = diff_usec / 1000000.0;
-  fprintf(fp, "%4.3f", diff_sec);
-}	
   
+  snprintf1(buf, size, "%4.3f ", diff_sec);
+}
+
+
+int killed_by(int status) {
+#ifdef WTERMSIG
+  if (WIFSIGNALED(status))
+    return WTERMSIG(status);
+#endif
+  return 0;
+}       
 
 /* change_working_directory() changes rlwrap's working directory to /proc/<command_pid>/cwd
    (on systems where this makes sense, like linux and Solaris) */
@@ -443,10 +461,10 @@ change_working_directory()
 
   snprintf0(slaves_working_directory, MAXPATHLEN, "?");
 
-  if (!initialized && command_pid > 0) { 	/* first time we're called after birth of child */
+  if (!initialized && command_pid > 0) {        /* first time we're called after birth of child */
     snprintf1(proc_pid_cwd, MAXPATHLEN , "/proc/%d/cwd", command_pid);
     initialized = TRUE;
-  }	
+  }     
   if (chdir(proc_pid_cwd) == 0) {
 #ifdef HAVE_READLINK
     linklen = readlink(proc_pid_cwd, slaves_working_directory, MAXPATHLEN);
@@ -455,7 +473,7 @@ change_working_directory()
     else
       snprintf1(slaves_working_directory, MAXPATHLEN, "? (%s)", strerror(errno));
 #endif /* HAVE_READLINK */
-  }		    
+  }                 
 #else
   /* do nothing at all */
 #endif
@@ -468,9 +486,10 @@ change_working_directory()
 void log_terminal_settings(struct termios *terminal_settings) {
   if (!terminal_settings)
     return;
-  DPRINTF1(DEBUG_TERMIO, "clflag.ISIG is %s", isset(terminal_settings->c_cflag | ISIG));
-  DPRINTF1(DEBUG_TERMIO, "cc_c[VINTR] is %d", terminal_settings->c_cc[VINTR]);
-  DPRINTF1(DEBUG_TERMIO, "cc_c[VEOF] is %d", terminal_settings->c_cc[VEOF]);
+  DPRINTF3(DEBUG_TERMIO, "clflag.ISIG: %s, cc_c[VINTR]=%d, cc_c[VEOF]=%d",
+           isset(terminal_settings->c_cflag | ISIG),
+           terminal_settings->c_cc[VINTR],
+           terminal_settings->c_cc[VEOF]);
 }
 
 void log_fd_info(int fd) {
@@ -480,25 +499,17 @@ void log_fd_info(int fd) {
       DPRINTF1(DEBUG_TERMIO, "could not get terminal settings for fd %d", fd);
     } else {
       DPRINTF1(DEBUG_TERMIO, "terminal settings for fd %d:", fd);
-    }	
+    }   
     log_terminal_settings(&terminal_settings);
   }
-}	
+}       
 
 
 /* some last-minute checks before we can start */
 void
 last_minute_checks()
 {
-  if (rl_editing_mode == 0 && term_cursor_up == NULL)
-    mywarn("\nWith this terminal (%s) %s will print all input lines twice in vi mode.\n"
-	   "Try a different terminal (or try setenv TERM=ansi) to get rid of this!\n", term_name, program_name); 
-  #ifdef DEBUG
-  if (debug & DEBUG_MEMORY_MANAGEMENT)
-    mywarn("\nDebugging memory management uses a\n"
-	   "primitive and clunky home-grown malloc/free debugger which will leak \n"
-	   "lots of memory; it may also crash easily");
-  #endif
+
 }
 
 
@@ -511,7 +522,7 @@ void mymicrosleep(int msec) {
   timeout.tv_sec = sec;
   timeout.tv_usec =  1000 * msec;
   select (0,NULL,NULL,NULL,&timeout);
-}	
+}       
 
 
 
@@ -536,7 +547,7 @@ static void print_option(char shortopt, char *longopt, char*argument, int option
 #endif
 
   if (argument) {
-    maybe_optional = (optional_args && optional ? add3strings("[", argument,"]") :  add3strings(" <", argument,">"));
+    maybe_optional = (optional_args && optional ? add3strings("[", argument,"]") :  add3strings("  <", argument,">"));
     longoptional = (optional ? add3strings("[=", argument,"]") : add3strings("=<", argument, ">"));
   }
   
@@ -560,8 +571,8 @@ void
 usage(int status)
 {
   fprintf(stderr, "Usage: %s [options] command ...\n"
-	          "\n"
-	          "Options:\n", program_name);
+          "\n"
+          "Options:\n", program_name);
 
   print_option('a', "always-readline", "password:", TRUE, NULL);
   print_option('A', "ansi-colour-aware", NULL, FALSE, NULL);
@@ -570,7 +581,7 @@ usage(int status)
   print_option('C', "command-name", "name|N", FALSE, NULL);
   print_option('D', "history-no-dupes", "0|1|2", FALSE, NULL);
   print_option('f', "file", "completion list", FALSE,NULL);
-  print_option('g', "forget-matching", "regex", FALSE,NULL);
+  print_option('g', "forget-matching", "regexp", FALSE,NULL);
   print_option('h', "help", NULL, FALSE, NULL);
   print_option('H', "history-filename", "file", FALSE, NULL);
   print_option('i', "case-insensitive", NULL, FALSE, NULL);
@@ -578,7 +589,9 @@ usage(int status)
   print_option('l', "logfile", "file", FALSE, NULL);
   print_option('n', "no-warnings", NULL, FALSE, NULL);
   print_option('N', "no-children", NULL, FALSE, NULL);
-  print_option('p', "prompt-colour", "ANSI colour spec", TRUE, NULL);
+  print_option('o', "one-shot", NULL, FALSE, NULL);
+  print_option('O', "only-cook", "regexp", FALSE, NULL);
+  print_option('p', "prompt-colour", "colour", TRUE, NULL);
   print_option('P', "pre-given","input", FALSE, NULL);
   print_option('q', "quote-characters", "chars", FALSE, NULL);
   print_option('m', "multi-line", "newline substitute", TRUE, NULL);
@@ -596,32 +609,33 @@ usage(int status)
   print_option('T', "test-terminal", NULL, FALSE, NULL);
   print_option('d', "debug", "mask", TRUE, add3strings("(output sent to ", DEBUG_FILENAME,")"));
   fprintf(stderr,
-	  "             \n"
-	  "The -d or --debug option *must* come first\n"
-	  "The debugging mask is a bit mask obtained by adding:\n");
+          "             \n"
+          "The -d or --debug option *must* come first\n"
+          "The debugging mask is a bit mask obtained by adding:\n");
 
   print_debug_flag (DEBUG_TERMIO, "to debug termio,");
   print_debug_flag (DEBUG_SIGNALS, "signal handling,");
   print_debug_flag (DEBUG_READLINE, "readline,");
   print_debug_flag (DEBUG_MEMORY_MANAGEMENT, "memory management,");
   print_debug_flag (DEBUG_FILTERING, "and filtering.");
-  print_debug_flag (DEBUG_AD_HOC, "to see your own DEBUG_AD_HOC results");
+  /*  print_debug_flag (DEBUG_AD_HOC, "to see your own DEBUG_AD_HOC results"); */
   print_debug_flag (DEBUG_WITH_TIMESTAMPS, "to add (relative) timestamps");
   print_debug_flag (FORCE_HOMEGROWN_REDISPLAY, "to force the use of my_homegrown_redisplay()");
-  print_debug_flag (DEBUG_LONG_STRINGS, "to not limit the length of strings in debug log");
-  print_debug_flag (DEBUG_RACES, "add random delays to uncover race conditions");
-  fprintf(stderr,  "    default debug mask = %d (debug termio, signals and readline handling)\n",
-	  DEBUG_ALL);
+  print_debug_flag (DEBUG_LONG_STRINGS, "to not limit the length of strings in debug log (sloooow!)");
+  print_debug_flag (DEBUG_RACES, "add random delays to expose race conditions");
+  fprintf(stderr,  "    default debug mask = %d (debug termio, signals and readline handling)\n"
+                   "    use the shell construct $[ ] to calculate the mask, e.g. -d$[%d+%d+%d]\n",
+          DEBUG_DEFAULT, DEBUG_DEFAULT, DEBUG_WITH_TIMESTAMPS, DEBUG_RACES);
   
 #endif
 
   fprintf(stderr,
-	  "\n"
-	  "bug reports, suggestions, updates:\n"
-	  "http://utopia.knoware.nl/~hlub/uck/rlwrap/\n");
+          "\n"
+          "bug reports, suggestions, updates:\n"
+          "http://utopia.knoware.nl/~hlub/uck/rlwrap/\n");
 
   exit(status);
-}	    
+}           
   
 
 #ifdef DEBUG
@@ -631,14 +645,14 @@ usage(int status)
 /* malloc with simplistic error handling: just bail out when out of memory */
 void *
 mymalloc(size_t size)
-{			
+{                       
   void *ptr;
   ptr = malloc(size);
   if (ptr == NULL) {
     /* don't call myerror(), as this calls mymalloc() again */
     fprintf(stderr, "Out of memory: tried in vain to allocate %d bytes\n", (int) size);
     exit(EXIT_FAILURE);
-  }	
+  }     
   return ptr;
 }
 
@@ -646,13 +660,13 @@ mymalloc(size_t size)
 void myfree(void *ptr) {
   if (ptr)
     free(ptr);
-}	
+}       
   
 void mysetsid() {
 # ifdef HAVE_SETSID /* c'mon, this is POSIX! */
-    pid_t ret = setsid();
+  pid_t ret = setsid();
 
-    DPRINTF2(DEBUG_TERMIO, "setsid() returned %d %s", (int)ret,
-	     ERRMSG(ret < 0));
+  DPRINTF2(DEBUG_TERMIO, "setsid() returned %d %s", (int)ret,
+           ERRMSG(ret < 0));
 # endif
 }
